/** @file
 * @author darkside86-w00t
 *
 * The core implementation of the Luamod runtime and game engine interface.
 *
 * @todo a safe reimplementation of getfenv and setfenv for the user environment.
 */

#include "engine.h"
#include <string>
#include <cstring>

const uint MAX_TIMERS = 64; // more than enough, really

static lua_State* state = NULL;
static bool lockreload = false;

// these tables are made at startup and are used to fill up lua globals
// when lua is initialized
//
static hashtable<char*,lua_CFunction>* userapi = NULL;
static hashtable<char*,lua_CFunction>* protectedapi = NULL;
struct LuaClassDef 
{ 
	const char* name ; 
	lua_CFunction initfn; 
	const char* parent;
};
static vector<LuaClassDef>* userclasses = NULL;
static vector<LuaClassDef>* protectedclasses = NULL;

// code is executed in two separate fenvs. even if function refs are
// duplicated, this can allow getfenv to exist in user space without
// fears of accessing the fenv of a protected function
//
#ifndef STANDALONE
static const char keyuserfenv = 'k';
#endif
static const char keyprotectedfenv = 'k';

// event hook list: array of {func, fenv}
//
static const char keyeventhandlers = 'k';

#ifndef STANDALONE
// timer list
//
static const char keytimerlist = 'k';

// addon list (server already has file io and doesn't need special ways to
// load code from files). this is a hashtable of addonname-datatable pairs.
static const char keyaddonsloaded = 'k';

static const char keyaddonlock = 'k';
static const char keyaddonlist = 'k';

// array of tables being serialized... to prevent infinite recursion
//
static const char keysavedtablelist = 'k';
#endif

// server and client have different events
#ifndef STANDALONE
/// List of client-related event hooks.
const char* clientluaevents[] =
{
	"FINALIZE", // [when Lua environment itself is cleaned up]
    "QUIT", // [when game is quit]
	"LUARELOAD",
    "KEYPRESS", // sym, bdown
    "MOUSEMOTION", // x, y, xrel, yrel
    "MOUSEBUTTON", // btn, state, x, y
    "CONSOLEMESSAGE", // text
    "MUSICDONE",
    "N_SERVINFO", // mycn, prot, sessionid
    "N_WELCOME", // hasmap
    "N_PAUSEGAME", // bPaused
    "N_TEXT", // d, text
    "N_SOUND", // d, soundnum
    "N_SAYTEAM", // d, text
    "N_DAMAGE", // target, actor, damage, armour, health
    "N_DIED", // victim, actor
    "N_EDITMSG", // d, msgtype
    "N_GUNSELECT", // d, gun
    "N_MAPCHANGE", // mapname
    "N_FORCEDEATH", // d
    "N_SPECTATOR", // d, bval
    "N_SETTEAM", // d, text
    "N_ANNOUNCE", // itemtype
    "N_NEWMAP", // d, size, newsize
    "N_REQAUTH", // text
    "N_INITAI", // bn,on,at,sk,pm,name,team
    "N_TAUNT", // d
    "N_ITEMLIST", 
    "N_ITEMSPAWN", // text, i
    "N_ITEMACC", // d, i
    "N_REMIP", // d
    "N_EDITVAR", // d, name
    "N_SERVMSG", // text
    "N_CONNECTED", // text
    "N_CDIS", // cn
    "N_MAPRELOAD", // map, nextmode
    "N_SWITCHNAME", // d, name
    "N_SWITCHMODEL", // d, model
    "N_SPAWN", // d
    "N_SPAWNSTATE", // s, life, gunselect
    "N_SHOTFX", // s, gun
	"N_EXPLODEFX", // e, gun, id
    "N_HITPUSH", // target, gun, damage
	"N_RESUME",
    "N_EDITMODE", // d, bval
    "N_EDITENT",
    "N_TIMEUP", // i
    "N_SENDDEMOLIST", // ndemos
    "N_DEMOPLAYBACK", // on, alias
    "N_CURRENTMASTER", // mn, priv
    "N_AUTHCHAL", // text
    "N_BASEINFO", // base,owner,enemy,converted,ammo
    "N_BASEREGEN", // d, health, armour, ammotype, ammo
    "N_BASES", // i,ammotype,owner,enemy,converted,ammo
    "N_BASESCORE", // base,text,total
    "N_REPAMMO", // r, ammotype
    "SHOOT", // d, gun
    "TARGETPLAYER", // d or nil
    "TARGETMONSTER", // m
    "TARGETNONE", // (no args)
};
#endif

/// List of server event hooks.
const char* serverluaevents[] =
{
	"DISCONNECTCLIENT", // hostname, reason
};

lua_State* getluastate()
{
	return state;
}

void getprotectedfenv()
{
	lua_pushlightuserdata(state, (void*)&keyprotectedfenv);
	lua_gettable(state, LUA_REGISTRYINDEX);
}

#ifndef STANDALONE
void getuserfenv()
{
	lua_pushlightuserdata(state, (void*)&keyuserfenv);
	lua_gettable(state, LUA_REGISTRYINDEX);
}
#endif

static void initprotectedfuncs()
{
	// conoutf(CON_ERROR, "initalizing protected api");

	// add the full stdlib
	//
	luaL_openlibs(state);

	// add user api functions
	//
	if(userapi)
	{
		enumeratekt(*userapi, char*, funcname, lua_CFunction, funcptr,
			// conoutf(" > adding user api function %s", funcname);
			lua_pushcfunction(state, funcptr);
			lua_setglobal(state, funcname);
		);
	}

	// add the protected api functions
	//
	if(protectedapi)
	{
		enumeratekt(*protectedapi, char*, funcname, lua_CFunction, funcptr,
			lua_pushcfunction(state, funcptr);
			lua_setglobal(state, funcname);
		);
	}

	// add the classes
	//
	if(userclasses)
	{
		loopv(*userclasses)
		{
			lua_pushcfunction(state, (*userclasses)[i].initfn);
			lua_call(state, 0, 0);
		}
	}

	if(protectedclasses)
	{
		loopv(*protectedclasses)
		{
			lua_pushcfunction(state, (*protectedclasses)[i].initfn);
			lua_call(state, 0, 0);
		}
	}
}

#ifndef STANDALONE
// can't just import the table, we have to make a copy in case the unsafe code
// tries to modify it
//
static void importuserlib(const char* libname, const char** funcnames, int nfuncs)
{
	lua_getglobal(state, libname); // ufenv, lib
	lua_pushstring(state, libname); // ufenv, lib, name,
	lua_newtable(state); // ufenv, lib, name, libcpy

	for(int i=0; i < nfuncs; ++i)
	{
		lua_pushstring(state, funcnames[i]); // ufenv, lib, name, libcpy, fn
		lua_pushvalue(state, -1); 			// ufenv, lib, name, libcpy, fn, fn,
		lua_gettable(state, -5); 		   // ufenv, lib, name, libcpy, fn, f
		lua_settable(state, -3); 		  // ufenv, lib, name, libcpy,
	}
	lua_settable(state, -4); // ufenv, lib,
	lua_pop(state, 1); // ufenv
}

// can't just import the table, we have to make a copy in case the unsafe code
// tries to modify it
//
static void importuserlibfull(const char* libname)
{
	lua_pushstring(state, libname); // ufenv, sz
	lua_newtable(state); // ufenv, sz, cpy

	lua_getglobal(state, libname);  // ufenv, sz, cpy, lib

	// if the lib has a metatable just copy the metatable
	//
	if(lua_getmetatable(state, -1) != 0)
	{
		lua_setmetatable(state, -3);
	}

	lua_pushnil(state); 		    // ufenv, sz, cpy, lib, nil
	while(lua_next(state, -2) != 0) // ufenv, sz, cpy, lib, key, val
	{
		lua_pushvalue(state, -2);  // ... key, val, key
		lua_pushvalue(state, -2); // ... key, val, key, val
		lua_settable(state, -6); // ... key, val
		lua_pop(state, 1);
	}
	lua_pop(state, 1); // ufenv, sz, cpy

	lua_settable(state, -3); // ufenv,
}

static void inituserfuncs()
{
	// conoutf(CON_ERROR, "initalizing user api");

	getuserfenv();

	// base lib imports
	//
	const char* imports[] =
	{
		"assert", "error", "getmetatable", "ipairs", "next", "pairs", "pcall",
			"loadstring", "print", "rawequal", "rawget", "rawset", "select",
			"setmetatable", "tonumber", "tostring", "type", "unpack",
			"_VERSION", "xpcall",
		// misc functions not in stdlib
		"typeof", "isobject",
	};

	for(uint i = 0; i < sizeof(imports) / sizeof(char*); ++i)
	{
		lua_pushstring(state, imports[i]);
		lua_getglobal(state, imports[i]);
		lua_settable(state, -3);
	}

	// set _G to be the user f-env for convenience
	lua_pushstring(state, "_G");
	getuserfenv();
	lua_settable(state, -3);

	// library table imports
	//
	importuserlibfull("coroutine");
	importuserlibfull("math");
	importuserlibfull("string");
	importuserlibfull("table");
	const char* os_imports[] =
	{
		"clock", "date", "difftime", "time"
	};
	importuserlib("os", os_imports, sizeof(os_imports) / sizeof(char*));

	if(userclasses)
	{
		/*enumeratekt(*userclasses, char*, name, lua_CFunction, fn,
			importuserlibfull(name);
		);*/
		loopv(*userclasses)
		{
			//printf("import: [%d] %s\n", i, (*userclasses)[i].name);
			importuserlibfull((*userclasses)[i].name);
		}
	}

	// add user api functions
	//
	if(userapi)
	{
		enumeratekt(*userapi, char*, funcname, lua_CFunction, funcptr,
			lua_pushstring(state, funcname);
			lua_pushcfunction(state, funcptr);
			lua_settable(state, -3);
		);
	}

	lua_pop(state,1);
}
#endif

static void initluaeventtable()
{
	lua_pushlightuserdata(state, (void*)&keyeventhandlers);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);

	lua_pushlightuserdata(state, (void*)&keyeventhandlers);
	lua_gettable(state, LUA_REGISTRYINDEX);

#ifndef STANDALONE
	for(uint i=0; i < sizeof(clientluaevents) / sizeof(char*); ++i)
	{
		lua_pushstring(state, clientluaevents[i]);
		lua_newtable(state);
		lua_settable(state, -3);
	}
#endif
	for(uint i=0; i < sizeof(serverluaevents) / sizeof(char*); ++i)
	{
		lua_pushstring(state, serverluaevents[i]);
		lua_newtable(state);
		lua_settable(state, -3);
	}
	lua_pop(state, 1);
}

#ifndef STANDALONE
static void inittimertable()
{
	lua_pushlightuserdata(state, (void*)&keytimerlist);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);
}

// ParseTocFile(filename) : table
//--------------------------------
// Parses an addon's table of contents file and returns a table with the
// fields: author, luafiles, savedvars.
//
const char TocParseLib[] =
{
#include "parsetocfile.lua.h"
};

void initsavedluavars(const char* name)
{
	// should be in [homesettings]/SavedVariables/addons/$name.lua
	//
	extern string homedir;
	std::string dirname = std::string(homedir) + "SavedVariables";
	vector<char*> dc;
	if( !fileexists(dirname.c_str(), "r") )
	{
		createdir(dirname.c_str()); // folders work differently than files
	}

	dirname = dirname + "/addons";

	if( !fileexists(dirname.c_str(),"r") )
	{
		createdir(dirname.c_str());
	}

	// std::string relname = std::string("SavedVariables") + "/addons/" + name + ".lua";
	dirname  = dirname + "/" + name + ".lua";

	if( fileexists(dirname.c_str(), "r") )
	{
		int notOk = luaL_loadfile(state, dirname.c_str());
		if(notOk)
		{
			conoutf(CON_ERROR, "Failed to load configuration file for addon \"%s\"", name);
			lua_pop(state, 1);
			return;
		}
		getuserfenv();
		lua_setfenv(state, -2);
		notOk = lua_pcall(state, 0, 0, 0);
		if(notOk)
		{
			conoutf(CON_ERROR, "Failed to load configuration file for \"%s\"", name);
			lua_pop(state, 1);
			return;
		}
	}
}

// table at -1, varname at -2, writes tables recursively. only strings
// numbers, and tables are considered.
//
static void writetable(stream* fileout, int level)
{
	lua_pushnil(state);
	char buf[512];
	while(lua_next(state, -2))
	{
		int valtype = lua_type(state, -1);
		if(valtype==LUA_TSTRING||valtype==LUA_TNUMBER||valtype==LUA_TTABLE||valtype==LUA_TBOOLEAN)
		{
			loopi(level) { fileout->putstring("\t"); }

			if(lua_type(state, -2) == LUA_TSTRING)
			{
				fileout->putstring("[\"");
				fileout->putstring(lua_tostring(state, -2));
				fileout->putstring("\"] = ");
			}
			else if(lua_type(state, -2) == LUA_TNUMBER)
			{
				// array indices do not need to be written
				// handle decimal vs integer somehow???
				if( lua_tointeger(state, -2) != lua_tonumber(state, -2) )
				{
					formatstring(buf)("[%f] = ", lua_tonumber(state, -2));
					fileout->putstring(buf);
				}
			}

			switch(lua_type(state, -1))
			{
				case LUA_TNUMBER:
					lua_getglobal(state, "tostring");
					lua_pushvalue(state, -2);
					lua_call(state, 1, 1);
					fileout->putstring(lua_tostring(state, -1));
					lua_pop(state, 1);
					break;
				case LUA_TSTRING:
					lua_getglobal(state, "string");
					lua_getfield(state, -1, "format");
					lua_replace(state, -2);
					lua_pushstring(state, "%q");
					lua_pushvalue(state, -3);
					lua_call(state, 2, 1);
					fileout->putstring(lua_tostring(state, -1));
					lua_pop(state, 1);
					break;
				case LUA_TBOOLEAN:
					fileout->putstring(lua_toboolean(state, -1) != 0? "true" : "false");
					break;
				case LUA_TTABLE:
					fileout->putline("{");
					// this table cannot be in keysavedtablelist
					//
					lua_pushlightuserdata(state, (void*)&keysavedtablelist);
					lua_gettable(state, LUA_REGISTRYINDEX);
					lua_pushvalue(state, -2); // table to be written
					lua_gettable(state, -2); // is it a key?
					bool okTable = lua_isnil(state, -1);
					lua_pop(state, 2);
					if(okTable)
					{
						lua_pushlightuserdata(state, (void*)&keysavedtablelist);
						lua_gettable(state, LUA_REGISTRYINDEX);
						lua_pushvalue(state, -2); // table to be written
						lua_pushboolean(state, 1);
						lua_settable(state, -3);
						lua_pop(state, 1);
						writetable(fileout, level+1);
					}
					else printf("SavedVariables: ignoring a table\n");

					loopi(level) { fileout->putstring("\t"); }
					fileout->putstring("}");
					break;
			}
			fileout->putline(",");
		}
		lua_pop(state, 1);
	}
}

void writeaddonconfig(char* name)
{
	// load the data table from registry
	//
	lua_pushlightuserdata(state, (void*)&keyaddonsloaded);
	lua_gettable(state, LUA_REGISTRYINDEX);
	lua_pushstring(state, name);
	lua_gettable(state, -2);
	lua_replace(state, -2);
	if(lua_isnumber(state, -1))
	{
		conoutf(CON_ERROR, "Error: %s is a number not an addon table", name);
		lua_pop(state, 1);
		return;
	}
	if(lua_isnil(state, -1))
	{
		// no addon with this name
		//
		conoutf(CON_ERROR, "Warning: addon %s isn't loaded", name);
		lua_pop(state, 1);
		return;
	}

	lua_pushstring(state, "savedvars");
	lua_gettable(state, -2);
	lua_replace(state, -2);
	if(lua_isnil(state, -1))
	{
		// no saved vars
		//
		lua_pop(state, 1);
		return;
	}

	std::string configname = std::string("SavedVariables/addons/") + name + ".lua";
	stream* configfile = openfile(configname.c_str(), "w");
	if(configfile == NULL)
	{
		conoutf(CON_ERROR, "Failed to write addon config file for %s", name);
		lua_pop(state, 1);
		return;
	}

	lua_pushnil(state);
	while(lua_next(state, -2))
	{
		// i, varname
		getuserfenv(); // i, varname, usrfenv
		lua_pushvalue(state, -2); // i, varname, userfenv, varname,
		lua_gettable(state, -2); // i, varname, userfenv, value
		lua_replace(state, -2); // i, varname, value

		if(!lua_isnil(state, -1))
		{
			switch(lua_type(state, -1))
			{
				case LUA_TNUMBER:
					configfile->putstring(lua_tostring(state, -2));
					configfile->putstring(" = ");
					// formatstring(buf)("%g", lua_tonumber(state, -1));
					// configfile->putline(buf);
					lua_getglobal(state, "tostring");
					lua_pushvalue(state, -2);
					lua_call(state, 1, 1);
					configfile->putline(lua_tostring(state, -1));
					lua_pop(state, 1);
					break;

				case LUA_TSTRING:
					configfile->putstring(lua_tostring(state, -2));
					configfile->putstring(" = ");
					// have to use string.format('%q') on it
					//
					lua_getglobal(state, "string");
					lua_getfield(state, -1, "format");
					lua_replace(state, -2);
					lua_pushstring(state, "%q");
					lua_pushvalue(state, -3);
					lua_call(state, 2, 1);
					configfile->putline(lua_tostring(state, -1));
					lua_pop(state, 1);
					break;

				case LUA_TBOOLEAN:
					configfile->putstring(lua_tostring(state, -2));
					configfile->putstring(" = ");
					configfile->putline(lua_toboolean(state, -1) != 0 ? "true" : "false");
					break;

				case LUA_TTABLE:
					configfile->putstring(lua_tostring(state, -2));
					configfile->putline(" = {");

					// this table cannot be in keysavedtablelist
					//
					lua_pushlightuserdata(state, (void*)&keysavedtablelist);
					lua_gettable(state, LUA_REGISTRYINDEX);
					lua_pushvalue(state, -2); // table to be written
					lua_gettable(state, -2); // is it a key?
					bool okTable = lua_isnil(state, -1);
					lua_pop(state, 2);

					if(okTable)
					{
						lua_pushlightuserdata(state, (void*)&keysavedtablelist);
						lua_gettable(state, LUA_REGISTRYINDEX);
						lua_pushvalue(state, -2); // table to be written
						lua_pushboolean(state, 1);
						lua_settable(state, -3);
						lua_pop(state, 1);
						writetable(configfile, 1);
					}
					else printf("SavedVariables: ignoring a recursive table\n");

					configfile->putline("}");
					break;
			}
		}

		lua_pop(state, 2); // fenv value and current value
	}

	configfile->close();
	delete configfile;
	conoutf("Successfully wrote configuration for %s", name);
	lua_pop(state, 1);
}
// COMMANDN(saveaddon, writeaddonconfig,"s");

static void saveaddonsconfig()
{
	// first clear table list
	//
	lua_pushlightuserdata(state, (void*)&keysavedtablelist);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(state, (void*)&keyaddonsloaded);
    lua_gettable(state, LUA_REGISTRYINDEX);
    lua_pushnil(state);
    while(lua_next(state, -2) != 0)
    {
        writeaddonconfig((char*)lua_tostring(state, -2));
        lua_pop(state, 1);
    }
    lua_pop(state,1);

    // write disable addons to addons.lua if any
    //
    getuserfenv();
    lua_pushstring(state, "DisabledAddons");
    lua_gettable(state, -2);
    lua_replace(state, -2);
    if(lua_isnil(state, -1))
    {
		lua_pop(state, 1);
		lua_newtable(state);
    }
    stream* addonCfg = openfile("addons.lua", "w");

    if(!addonCfg) { lua_pop(state, 1) ; return; }

    addonCfg->putline("DisabledAddons = {");

    lua_pushnil(state);
    while(lua_next(state, -2))
    {
    	std::string line = std::string("\"") + lua_tostring(state,-1) + "\",";
    	addonCfg->putline(line.c_str());
    	lua_pop(state, 1);
    }
    addonCfg->putline("}");
    addonCfg->close();
    delete addonCfg;
    lua_pop(state, 1);
}

#define ADDON_LOAD_FAIL(STACK) { \
	lua_pushlightuserdata(L, (void*)&keyaddonsloaded); \
	lua_gettable(L, LUA_REGISTRYINDEX); lua_pushstring(L, addon); \
	lua_pushnumber(L, -1); lua_settable(L, -3); lua_pop(L, STACK+1); \
	return 0; \
}

int loadaddon(lua_State* L)
{
	const char* addon = luaL_checkstring(L, 1);
	// is it locked?
	lua_pushlightuserdata(L, (void*)&keyaddonlock);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_gettable(L, -2);
	bool isLocked = !lua_isnil(L, -1);
	lua_pop(L, 2);
	if(isLocked)
	{
		conoutf(CON_ERROR, "%s is busy right now", addon);
		return 0;
	}
	// failed load or not in list? return false
	lua_pushlightuserdata(L, (void*)&keyaddonlist);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_gettable(L, -2);
	bool inList = !lua_isnil(L, -1);
	lua_pop(L, 2);
	lua_pushlightuserdata(L, (void*)&keyaddonsloaded);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_gettable(L, -2);
	bool failedLoad = false;
	bool alreadyLoaded = false;
	if (lua_isnumber(L, -1)) failedLoad = lua_tonumber(L,-1) == -1;
	if (lua_toboolean(L, -1)!=0) alreadyLoaded = true;
	lua_pop(L, 2);
	if(failedLoad || !inList)
	{
		return 0;
	}
	if(alreadyLoaded)
	{
		lua_pushboolean(L, 1);
		return 1;
	}
	// needs to be loaded, so read the toc file.
	vector<char*> dircontents;
	std::string addondir = std::string("addons/") + addon;
	bool isFolder = listdir(addondir.c_str(), NULL, dircontents);
	if(!isFolder)
	{
		conoutf(CON_ERROR, "%s is not a valid addon", addon);
		ADDON_LOAD_FAIL(0);
	}
	std::string tocfname = addondir + "/" + addon + ".toc";
	stream* tocfile = openfile(tocfname.c_str(), "r");
	if(!tocfile)
	{
		conoutf(CON_ERROR, "Failed to read %s", tocfname.c_str());
		ADDON_LOAD_FAIL(0);
	}
	tocfile->close();
	delete tocfile;
	int luaerr = luaL_loadbuffer(L, TocParseLib, sizeof(TocParseLib)-1, "parsetocfile.lua");
	if( luaerr != 0 )
	{
		conoutf(CON_ERROR, "Unable to open addon: %s", lua_tostring(L, -1));
		ADDON_LOAD_FAIL(1);
	}
	luaerr = lua_pcall(L, 0, 0, 0);
	if( luaerr != 0 )
	{
		conoutf(CON_ERROR, "Unable to open addon: %s", lua_tostring(L, -1));
		ADDON_LOAD_FAIL(1);
	}
	lua_getglobal(L, "ParseTocFile");
	lua_pushstring(L, tocfname.c_str());
	luaerr = lua_pcall(L, 1, 1, 0);
	if( luaerr != 0 )
	{
		conoutf(CON_ERROR, "Unable to read addon: %s", lua_tostring(L, -1));
		ADDON_LOAD_FAIL(1);
	}
	// data table is now on stack at -1
	// if there are any strings in data.depends we must try to resolve them
	lua_pushstring(L, "depends");
	lua_gettable(L, -2);
	if(!lua_isnil(L, -1))
	{
		lua_pushnil(L);
		while(lua_next(L, -2))
		{	// data, i, name
			lua_pushlightuserdata(L, (void*)&keyaddonlock);
			lua_gettable(L, LUA_REGISTRYINDEX);
			lua_pushstring(L, addon);
			lua_pushboolean(L, 1);
			lua_settable(L, -3); // data, i, name, lock
			lua_pushcfunction(L, loadaddon); // data, i, name, lock, f
			lua_pushvalue(L, -3);
			lua_call(L, 1, 1); // data, i, name, lock, t|f
			bool depOk = lua_toboolean(L, -1) != 0;
			lua_pop(L, 1);
			lua_pushstring(L, addon);
			lua_pushnil(L);
			lua_settable(L, -3);
			lua_pop(L, 1); // remove lock from stack
			if(!depOk)
			{
				conoutf(CON_ERROR, "Addon %s failed to load: missing dependency %s", addon, lua_tostring(L, -1));
				ADDON_LOAD_FAIL(2);
			}
			lua_pop(L, 1);
		}
	}
	lua_pop(L, 1);
	// parse the addon's cfg and lua files
	initsavedluavars(addon);
	lua_getfield(L, -1, "luafiles"); // data luafiles
	lua_pushnil(L);
	while(lua_next(L, -2)) // data luafiles k v
	{
		lua_pushstring(L, addondir.c_str());
		lua_pushstring(L, "/");
		lua_pushvalue(L, -3);
		lua_concat(L, 3);
		luaerr = luaL_loadfile(L, lua_tostring(L,-1));
		lua_replace(L, -2); // replace the string at -2 with whatever this is
		if ( luaerr != 0 )
		{
			conoutf(CON_ERROR, "Addon loading error: %s", lua_tostring(L, -1));
			ADDON_LOAD_FAIL(5);
		}
		getuserfenv();
		lua_setfenv(L, -2);
		luaerr = lua_pcall(L, 0, 0, 0);
		if ( luaerr != 0 )
		{
			conoutf(CON_ERROR, "Addon run error: %s", lua_tostring(L, -1));
			ADDON_LOAD_FAIL(5);
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1); // remove luafiles from stack
	lua_getfield(L, -1, "cfgfiles");
	lua_pushnil(L);
	while(lua_next(L, -2))
	{
		lua_pushstring(L, addondir.c_str());
		lua_pushstring(L, "/");
		lua_pushvalue(L, -3);
		lua_concat(L, 3);
		bool cfgok = execfile(lua_tostring(L,-1), true);
		lua_pop(L, 1);
		if( !cfgok )
		{
			conoutf(CON_ERROR, "Addon cfg file error");
			ADDON_LOAD_FAIL(4);
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
	conoutf("Successfully loaded addon %s", addon);
	lua_pushlightuserdata(L, (void*)&keyaddonsloaded);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushstring(L, addon); // loaded, addon
	lua_pushvalue(L, -3); // loaded, addon, data
	lua_settable(L, -3);
	lua_pushboolean(L, 1);
	lua_replace(L, -2);
	return 1;
}

void loadalladdons()
{
	// create tables
	lua_pushlightuserdata(state, (void*)&keyaddonlist);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);
	lua_pushlightuserdata(state, (void*)&keyaddonlock);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);
	lua_pushlightuserdata(state, (void*)&keyaddonsloaded);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);
	
	// fill addon string array from directory contents and disabled addon info
	extern string homedir;
	vector<char*> listing;
	bool ok = listdir("addons", NULL, listing);
	if (!ok)
	{
		conoutf(CON_ERROR, "Unable to read addon folder.");
		return;
	}
	std::string addonFile = std::string(homedir) + "addons.lua";
	if(!fileexists(addonFile.c_str(), "r"))
	{
		stream* addonsLua = openfile("addons.lua", "w");
		if(!addonsLua)
		{
			conoutf(CON_ERROR, "Unable to read addon list.");
			return;
		}
		addonsLua->putline("DisabledAddons = {}");
		addonsLua->close();
		delete addonsLua;
	}
	int luaerror = luaL_loadfile(state, addonFile.c_str());
	if(luaerror != 0)
	{
		lua_pop(state, 1);
		conoutf("Unable to load addon list");
		return;
	}
	getuserfenv();
	lua_setfenv(state, -2);
	luaerror = lua_pcall(state, 0, 0, 0);
	if(luaerror != 0)
	{
		lua_pop(state, 1);
		conoutf(CON_ERROR, "Unable to parse addon list");
		return;
	}
	// build an assoc-array on stack of disabled addons.
	getuserfenv();
	lua_pushstring(state, "DisabledAddons");
	lua_gettable(state, -2); // usr da|nil
	lua_replace(state, -2); // da|nil
	if(lua_isnil(state, -1))
	{
		lua_pop(state, 1);
		lua_newtable(state);
	} // da
	lua_newtable(state); // da, nt
	lua_pushnil(state); // da, nt, nil
	while(lua_next(state, -3)) // da, nt, k, addonname
	{
		lua_pushboolean(state, 1); // da, nt, k, addonname, true
		lua_settable(state, -4); // da, nt, k
	} // da, nt
	lua_replace(state, -2); // nt
	// fill addonlist with valid strings
	lua_pushlightuserdata(state, (void*)&keyaddonlist);
	lua_gettable(state, LUA_REGISTRYINDEX); // nt list
	loopv(listing)
	{
		lua_pushstring(state, listing[i]); //nt list key
		lua_gettable(state, -3); // nt list [true|nil]
		if(lua_isnil(state, -1) != 0)
		{
			// if a '.' is found in this name, it can't be a valid addon
			//
			lua_getglobal(state, "string");
			lua_getfield(state, -1, "find");
			lua_replace(state, -2); // ... find (
			lua_pushstring(state, listing[i]); // ... find(addon, 
			lua_pushstring(state, "%."); // ... find(addon, "%.")
			lua_call(state, 2,1); // nt list nil findresult
			if(lua_isnil(state,-1))
			{
				lua_pushstring(state, listing[i]); // nt list nil fr key
				lua_pushboolean(state, 1); // nt list nil fr key true
				lua_settable(state, -5); // nt list nil fr
			}
			lua_pop(state, 1); // nt list nil
		}
		lua_pop(state, 1); // nt list
	}
	lua_replace(state, -2); // list
	
	lua_pushnil(state);
	while(lua_next(state, -2))
	{
		lua_pushcfunction(state, loadaddon);
		lua_pushvalue(state, -3);
		lua_call(state, 1, 0); // ignore return value when called here
		lua_pop(state, 1);
	}
	lua_pop(state, 1);
	
	// remove -1 values from keyaddonsloaded
	lua_newtable(state); // store keys whose value are -1 here
	int i = 1;
	lua_pushlightuserdata(state, (void*)&keyaddonsloaded);
	lua_gettable(state, LUA_REGISTRYINDEX);
	lua_pushnil(state); // nt loaded nil
	while(lua_next(state, -2)) // nt loaded addon (data|-1)
	{
		if(lua_isnumber(state, -1))
		{
			if(lua_tonumber(state, -1) == -1)
			{
				lua_pushinteger(state, i++); // nt ld ad v i++
				lua_pushvalue(state, -3); // nt ld ad v i++ ad
				lua_settable(state, -6); // nt ld ad v
			}
		}
		lua_pop(state, 1);
	} // nt loaded
	
	lua_pushnil(state); // nt loaded nil
	while(lua_next(state, -3))
	{ // nt loaded i adn
		lua_pushvalue(state, -1); // nt loaded i adn adn
		// printf("I need to cleanup %s (-1)\n", lua_tostring(state, -1));
		lua_pushnil(state); // nt loaded i adn adn nil
		lua_settable(state, -5);
		lua_pop(state, 1);
	}
	lua_pop(state, 2);
}
#endif

bool initlua(bool initfile, bool addons, bool triggerevent)
{
	if(lockreload) return false;

	if(state)
	{
		cleanuplua();
	}

	state = lua_open();
	if(!state)
	{
		conoutf(CON_ERROR, "Call to lua_open failed");
		return false;
	}

	initluaeventtable();
#ifndef STANDALONE
	inittimertable();
#endif

	// create two separate fenvs and store them in the registry
	//
#ifndef STANDALONE
	lua_pushlightuserdata(state, (void*)&keyuserfenv);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);
#endif
	lua_pushlightuserdata(state, (void*)&keyprotectedfenv);
	lua_pushthread(state);
	lua_getfenv(state, -1);
	lua_replace(state, -2);
	lua_settable(state, LUA_REGISTRYINDEX);

	// initialize the contents of the protected api
	//
	initprotectedfuncs();
#ifndef STANDALONE
	inituserfuncs();
	
	extern int luaopen_luaui(lua_State*);
	lua_pushcfunction(state, luaopen_luaui);
	lua_call(state, 0, 0);

	lua_pushlightuserdata(state, (void*)&keyaddonsloaded);
	lua_newtable(state);
	lua_settable(state, LUA_REGISTRYINDEX);

	if(addons)
		loadalladdons();
#endif

	if(triggerevent)
		callluaevent("LUARELOAD", "");

	return true;
}

ICOMMAND(reloadlua, "", (), intret(initlua(true,true,true) ? 1 : 0));

void cleanuplua()
{
	if(state)
	{
		callluaevent("FINALIZE", "");
#ifndef STANDALONE
	    saveaddonsconfig();
#endif
		lua_close(state);
	}
	state = NULL;
}

int addluauserapi(const char* name, int (*func)(lua_State*))
{
	if(userapi == NULL) userapi = new hashtable<char*,lua_CFunction>();
	userapi->access((char*)name, func);
	return 0;
}

int addluaprotectedapi(const char* name, int (*func)(lua_State*))
{
	if(protectedapi == NULL) protectedapi = new hashtable<char*,lua_CFunction>();
	protectedapi->access((char*)name, func);
	return 0;
}

int addluaclass(bool prot, const char* name, int (*func)(lua_State*), const char* parent)
{
	vector<LuaClassDef>* classes;
	if(prot)
	{
		if(protectedclasses == NULL) protectedclasses = new vector<LuaClassDef>();
		classes = protectedclasses;
	}
	else
	{
		if(userclasses == NULL) userclasses = new vector<LuaClassDef>();
		classes = userclasses;
	}
	LuaClassDef def;
	def.name = name;
	def.initfn = func;
	def.parent = parent;
	loopv(*classes)
	{
		// is this out of order parent?
		if((*classes)[i].parent != NULL)
		{
			if(std::string((*classes)[i].parent) == std::string(name))
			{
				printf("Putting child %s after new %s\n", (*classes)[i].name, name);
				classes->insert(i, &def, 1);
				return -1;
			}
		}
	}
	classes->add(def);
	printf("Added %s to the end\n", (*classes)[classes->length()-1].name);
	return -1;
}

int addluabaseclass(bool prot, const char* name, int (*func)(lua_State*))
{
	vector<LuaClassDef>* classes;
	if(prot)
	{
		if(protectedclasses == NULL) protectedclasses = new vector<LuaClassDef>();
		classes = protectedclasses;
	}
	else
	{
		if(userclasses == NULL) userclasses = new vector<LuaClassDef>();
		classes = userclasses;
	}
	LuaClassDef def;
	def.name = name;
	def.initfn = func;
	def.parent = NULL;
	classes->insert(0, &def, 1);
	return -1;
}

/** 
 * Classes that send existing objects into Lua but delete
 * them before Lua is finalized need to add this to their destructor.
 * @param object should be the "this" pointer in the destructor
 * @param cls the class name as defined in Lua.
*/
void luadestroyobject(void* object, const char* cls)
{
    if(state)
    {
        classbind::removeobject(state, object, cls);
    }
}

/** 
 * A va_arg interface for the game to call Lua events anywhere without
 * knowledge of Lua header files. Args are formatted as follows,
 * 'i' or integer, 's' for c-string, 'c' for class-name, class-object
*/
void callluaevent(const char* evt, const char* fmt, ...)
{
    if(state==NULL) return;

    lua_State* L = state;
    int nargs = 0;
    int tblinc = 0; // used {} array formatting
    int* curinc = &nargs;

    // get va_arg and push them into Lua stack
    //
    va_list argptr;
    va_start(argptr,fmt);
    while(*fmt) // fmt string ends with '\0'
    {
        switch(*fmt)
        {
        case 's': // string
            {
            const char* str = va_arg(argptr, char*);
            lua_pushstring(L, str);
            ++(*curinc);
            break;
            }
        case 'i': // integer
            {
            int n = va_arg(argptr, int);
            lua_pushinteger(L, n);
            ++(*curinc);
            break;
            }
        case 'f': // float or double
            {
            double f = va_arg(argptr, double);
            lua_pushnumber(L, f);
            ++(*curinc);
            break;
            }
        case 'b': // bool
            {
            bool b = va_arg(argptr, int) != 0;
            lua_pushboolean(L, b);
            ++(*curinc);
            break;
            }
        case 'c': // class object
            {
            const char* clsname = va_arg(argptr, char*);
            void* object = va_arg(argptr, void*);
            classbind::pushobject(state, object, clsname);
            ++(*curinc);
            break;
            }
        case '{':
            {
            // if curinc is already pointing at the tbl incrementer, it is an
            // error... can't nest arrays or tables...
            if(curinc == &tblinc)
                luaL_error(L, "Unexpected '{' in callluaevent");
            curinc=&tblinc;
            *curinc = 0; // arrays start at 1 in lua
            lua_newtable(L);
            break;
            }
        case '}': // close an array started with {
            {
            if(curinc != &tblinc)
                luaL_error(L, "Unexpected '}' in callluaevent");
            // fill up the array backwards (elements were placed like that)
            for(int n=tblinc; n >= 1; --n)
            {
                lua_pushinteger(L, n);
                lua_insert(L, -2);
                lua_settable(L, -(n+2));
            }
            curinc = &nargs;
            ++(*curinc);
            break;
            }

        default:
            luaL_error(L, "Invalid type specifier `%c' in calluaevent", *fmt);
        }
        ++fmt;
    }
    va_end(argptr);

    // get event hook table
    //
    lua_pushlightuserdata(L, (void*)&keyeventhandlers);
    lua_gettable(L, LUA_REGISTRYINDEX);

    // get hook list for this function
    //
    lua_pushstring(L, evt);
    lua_gettable(L, -2);

    // get number of functions in table
    //
    int num = lua_objlen(L, -1);

    // remove hook register list
    //
    lua_remove(L, -2);

    if(num==0) // no events, cleanup and return
    {
        lua_pop(L,1+nargs);
        return;
    }
    else
    {
        for(int eachhook=1; eachhook <= num-1; ++eachhook)
        {
            lua_pushnumber(L, eachhook);
            lua_gettable(L, -2);

            // repush args for each but the last
            //
            for(int eachArg=1; eachArg <= nargs; ++eachArg)
                lua_pushvalue(L, -(nargs+2));
			lockreload = true;
			int err = lua_pcall(L, nargs, 0, 0);
			lockreload = false;
            if(err != 0)
            {
                conoutf(CON_ERROR, "%s: %s", evt, lua_tostring(L,-1));
                lua_pop(L, 2);
                return;
            }
        }

        // for the last func in the list, the args are already on the stack
        //
        lua_pushnumber(L, num);
        lua_gettable(L, -2);
        lua_remove(L, -2); // remove func list
        lua_insert(L, -(nargs+1)); // move function to bottom
		lockreload = true;
		int l_err = lua_pcall(L, nargs, 0, 0);
		lockreload = false;
        if(l_err !=0 )
        {
            conoutf("%s: %s", evt, lua_tostring(L,-1));
            lua_pop(L, 1); // remove error msg
            return;
        }
    }
    // nothing should be on stack now
}

#ifndef STANDALONE
void checkluatimers(int millis)
{
    if(!state) return;
    lua_State* L = state;

    lua_pushlightuserdata(L, (void*)&keytimerlist);
    lua_gettable(L, LUA_REGISTRYINDEX);
    lua_pushnil(L);
    int size = 0;
    while(lua_next(L,-2) != 0)
    {
        // value is at -1
        int key = lua_tointeger(L,-2); // l i t
        if(lua_type(L, -1) != LUA_TNUMBER)
        {
            lua_pushstring(L, "time"); // l i t s
            lua_gettable(L, -2); // l i t n
            int time = lua_tointeger(L, -1);
            lua_pop(L,1); // l i t
            if((millis) >= time)
            {
                // time to execute and remove this timer
                lua_pushstring(L, "func"); // l i t s
                lua_gettable(L, -2);       // l i t f
                // callable now on stack but first remove this item
                // (set to 0 not nil--if we are at end it will shorten the array causing error)
                lua_pushinteger(L, key); // l i t f k
                lua_pushinteger(L, 0);   // l i t f k N
                lua_settable(L, -6);     // l i t f
                int notOk = lua_pcall(L, 0, 0, 0); // l i t
                if(notOk) {
                    const char* reason = lua_tostring(L,-1);
                    conoutf(CON_ERROR, "Lua error: %s\n", reason);
                    lua_pop(L,1);
                }
            }
            // l i t
        }
        // l i t
        lua_pop(L,1); // remove value and use next key
        key = lua_tointeger(L, -1);
        size = lua_objlen(L, -2);
    }

    lua_pop(L,1); // remove table
}
#endif

/**
 * Executes a string of Lua code in the protected (default) environment.
 */
bool executeluaprotected(const char* code)
{
	if(!state) return false;

	int error = luaL_loadbuffer(state, code, strlen(code), "protected code") ||
				lua_pcall(state, 0, 0, 0);

	// errors will be handled differently in protected code...
	// this is for debugging purposes
	//
	if(error)
	{
		conoutf(CON_ERROR, "Luap error: %s", lua_tostring(state, -1));
		lua_pop(state, 1);
		return false;
	}

	return true;
}
// for debugging purposes only, a luap command
ICOMMAND(luap, "Z", (char* code), intret(executeluaprotected(code) ? 1 : 0));

#ifndef STANDALONE
/**
 * Executes a string of code in the limited user environment.
 */
bool executeluauser(const char* code)
{
	if(!state) return false;

	int error = luaL_loadbuffer(state, code, strlen(code), "user code");
	if(error)
	{
		if(error == LUA_ERRSYNTAX)
		{
			conoutf(CON_ERROR, "Lua syntax error: %s", lua_tostring(state, -1));
		}
		else if(error == LUA_ERRMEM)
		{
			conoutf(CON_ERROR, "Lua memory error: %s", lua_tostring(state, -1));
		}
		lua_pop(state, 1);
		return false;
	}

	getuserfenv(); // put the user f-environment on stack (retrieved from registry)
	lua_setfenv(state, -2); // set the loaded buffer func's fenv
	error = lua_pcall(state, 0, 0, 0);
	if(error)
	{
		conoutf(CON_ERROR, "Lua: %s", lua_tostring(state, -1));
		lua_pop(state, 1);
		return false;
	}

	return true;
}
ICOMMAND(lua, "Z", (char* code), intret(executeluauser(code) ? 1 : 0));
ICOMMAND(execlua, "s", (char* code), intret(executeluauser(code) ? 1 : 0));
#else
ICOMMAND(lua, "Z", (char* code), intret(executeluaprotected(code) ? 1 : 0));
ICOMMAND(execlua, "s", (char* code), intret(executeluaprotected(code) ? 1 : 0));
#endif

// the lua api to luamod
///////////////////////////////////////////////////////////////////////////////

/* AddEventHook(name, func)
 |--------------------------
 | Adds an event hook to be called by the various C++ functions in the game.
 | The environment in which the function is run depends on the environment
 | in which AddEventHook is called, of course.
*/
int lua_AddEventHook(lua_State* L)
{
	const char* eventname = luaL_checkstring(L, 1);
	if(lua_isnil(L, 2))
		luaL_typerror(L, 2, "callable object");

	lua_pushlightuserdata(L, (void*)&keyeventhandlers);
	lua_gettable(L, LUA_REGISTRYINDEX); // tbl

	lua_pushvalue(L, 1); // tbl str
	lua_gettable(L, -2); // tbl hooks
	lua_replace(L, -2); // hooks

	if(lua_isnil(L, -1))
		return luaL_error(L, "`%s' is not a valid event name", eventname);

	int nextindex = lua_objlen(L, -1);
	lua_pushinteger(L, ++nextindex); // hooks ++i
	lua_pushvalue(L, 2); // hooks i fn
	lua_settable(L, -3); // hooks

	lua_pop(L, 1); // remove event list from stack

	return 0;
}
ADDLUAUSERAPI(AddEventHook);

#ifndef STANDALONE
/* AddTimer(func, time)
 |----------------------
 | Add a function to be called after $time millis have passed.
*/
int lua_AddTimer(lua_State* L)
{
    int ftype = lua_type(L, 1);
    if(!(ftype == LUA_TFUNCTION || ftype == LUA_TUSERDATA || ftype == LUA_TTABLE))
        luaL_typerror(L, 1, "Callable object");
    int millis = luaL_checkinteger(L,2);

    // create a table {func=..., time=...} and put it in a list
    //
    lua_newtable(L);
    lua_pushstring(L, "func");
    lua_pushvalue(L, 1);
    lua_settable(L, -3);
    lua_pushstring(L, "time");
    lua_pushinteger(L, millis+lastmillis);
    lua_settable(L, -3); // t

    // search list for a nil or empty space
    //
    lua_pushlightuserdata(L, (void*)&keytimerlist);
    lua_gettable(L, LUA_REGISTRYINDEX); // t l
    int i = 0;
    lua_pushnil(L); // t l N
    while(lua_next(L, -2) != 0)
    {
        i = lua_tointeger(L,-2); // t l k v
        // if(lua_isnil(L, -1))
        // is nil or set to 0, it is an open space
        if( lua_type(L, -1) == LUA_TNUMBER || lua_type(L,-1) == LUA_TNIL)
        {
            // printf("Found a nil at %d\n", i);
            lua_pop(L, 1); // t l k
            lua_pushvalue(L, -3); // t l k t
            lua_settable(L, -3); // t l
            lua_pop(L,2);
            return 0;
        }
        lua_pop(L,1); // t l k
    }
    // t l
    // if we reach here that means we have to put it at the end of the table
    ++i;
    // printf("Found no nil, i is now %d\n", i);
    if((unsigned)i > MAX_TIMERS)
    {
        luaL_error(L, "AddTimer: failed to add timer, too many timers already");
    }
    else
    {
        lua_pushinteger(L, i); // t l i
        lua_pushvalue(L, -3); // t l i t
        lua_settable(L, -3); // t l
    }
    lua_pop(L,2);
    return 0;
}
ADDLUAUSERAPI(AddTimer);

int lua_StopAllTimers(lua_State* L)
{
	lua_pushlightuserdata(L, (void*)&keytimerlist);
	lua_newtable(L);
	lua_settable(L, LUA_REGISTRYINDEX);
	return 0;
}
ADDLUAPROTECTEDAPI(StopAllTimers);
#endif

/* GetLuaMemoryUsage() : number
 |---------------------------------
 | Returns the amount of memory consumed by Lua in KB.
*/
int lua_GetLuaMemoryUsage(lua_State* L)
{
	lua_getglobal(L, "collectgarbage");
	lua_pushstring(L, "count");
	lua_call(L, 1, 1);
	return 1;
}
ADDLUAUSERAPI(GetLuaMemoryUsage);

/* getvar(varname) : string
 |-----------------------------------------------------------------------------
 | Returns the value of an alias or variable in cubescript referenced by
 | $varname. The result is always a string. If numeric values are expected they
 | should be explicitly converted via tonumber().
*/
int lua_getvar(lua_State* L)
{
    const char* varName = luaL_checkstring(L, 1);

    if(strlen(varName) > 255)
        return 0;

    char val[266];
    sprintf(val, "result $%s", varName);
    lua_pushstring(L, executeret(val));
    return 1;
}
ADDLUAUSERAPI(getvar);

/* setvar(name,value,[isalias=false])
 |-----------------------------------------------------------------------------
 | Sets a cubescript variable $name to the value $value. This also works on
 | special control variables such as "hidehud" but only if the third parameter
 | is true.
*/
int lua_setvar(lua_State* L)
{
    const char* varName = luaL_checkstring(L, 1);
    const char* value = lua_tostring(L, 2);
    bool isAlias = lua_toboolean(L, 3) != 0;

    if(lua_type(L,2)==LUA_TBOOLEAN)
    {
        if(lua_toboolean(L, 2))
            value = "1";
        else
            value = "0";
    }
    else
    {
        value = lua_tostring(L,2);
    }

    if(value == NULL) // tostring() failed to produce value and will segfault
    {
        luaL_error(L, "invalid value specified for setvar");
        return 0;
    }

    if(strlen(varName) + strlen(value) >= 239)
    {
        luaL_error(L, "variable name is too long");
        return 0;
    }

    char exestr[256];
    if(!isAlias)
        sprintf(exestr, "%s = %s", varName, value);
    else
        sprintf(exestr, "%s \"%s\"", varName, value);
	lockreload = true;
    lua_pushstring(L, executeret(exestr));
    lockreload = false;
    return 1;
}
ADDLUAUSERAPI(setvar);

/* runscript(cubescriptcode) : string
 |------------------------------------
 | Attempts to execute a line of cubescript and return the result
*/
int lua_runscript(lua_State *L)
{
	const char* cubeCmd = lua_tostring(L, -1);

	// either return a string representing the value (numerical or not)
	// or nil for no result.
	//
	lockreload = true;
	char* ret = executeret(cubeCmd);
	lockreload = false;

	if(ret != NULL)
		lua_pushstring(L, ret);
	else
		lua_pushnil(L);

	DELETEA(ret);

	return 1;
}
ADDLUAUSERAPI(runscript);

#ifndef STANDALONE
int lua_EnvTest(lua_State* L)
{
	getuserfenv();
	lua_getglobal(L, "getfenv");
	lua_pushnumber(L, 0);
	lua_call(L, 1, 1);
	bool equals = lua_equal(L, -1, -2) != 0;
	// lua_pop(L, 2);
	conoutf("EnvTest: is this user env? %s", equals ? "yes" : "no");
	return 2;
}
ADDLUAUSERAPI(EnvTest);

// gets the addon table returned by parseTocFile
int lua_GetAddonInfo(lua_State* L)
{
	const char* name = luaL_checkstring(L,1);
	lua_pushlightuserdata(L, (void*)&keyaddonsloaded);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushstring(L, name);
	lua_gettable(L, -2);
	lua_replace(L, -2);
	return 1;
}
ADDLUAPROTECTEDAPI(GetAddonInfo);

// get the f-environment of the user API
int lua_GetfenvUser(lua_State* L)
{
    getuserfenv();
    return 1;
}
ADDLUAPROTECTEDAPI(GetfenvUser);
#endif

int lua_ProtectedTest(lua_State* L)
{
	conoutf("Calling ProtectedTest!");
	return 0;
}
ADDLUAPROTECTEDAPI(ProtectedTest);

// TODO: safe reimplementation of getfenv and setfenv
