/** 
 * @file
 * @author darkside86-w00t
 * @brief a Lua library written for Cube 2 Engine for rendering graphical 2D objects in C++ but
 * declaring and manipulating them from Lua.
 * @todo UIVideo, UIWorldFrame, UIRichText, UIImage, UIRoundFrame, UIDialog, (UIWidget classes),
 * UIAnimation
 */

#include "engine.h"
#include <vector>
#include <string>
#include <cstring>

// various utilities, could be moved to luamodutil.cpp or something
#define NEW_WEAK_LIST(L, mode) { \
	lua_newtable(L); \
	lua_newtable(L); \
	lua_pushstring(L, #mode); \
	lua_setfield(L, -2, "__mode"); \
	lua_setmetatable(L, -2); \
}

int lua_getNumList(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TTABLE);
	
	int size = 0;
	lua_pushnil(L);
	while(lua_next(L, 1))
	{
		size++;
		lua_pop(L, 1);
	}
	lua_pushinteger(L, size);
	return 1;
}

int getNumList(lua_State* L, int table)
{
	int size = 0;
	lua_pushvalue(L, table);
	lua_pushnil(L);
	while(lua_next(L, -2))
	{
		size++;
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
	return size;
}

struct UIObject;

static const char keyUiParent = 'k';
static const char keyShownDlgs = 'k';
static const char keyListeners = 'k';
static const char keyFocus = 'k'; // current object in focus
static const char keyDrag = 'k'; // current object being dragged
static const char keyButtons = 'k'; // all UIButton instances

/** This is the top level container for all declared UIObjects. It cannot be directly accessed
 * from Lua, but is automatically created and destroyed when Lua is started or ended respectively.
 * Any UIObject declared with no parent will automatically have this as their parent.
 */
UIObject* luaUIParent = NULL;

/** This is a solid white 1x1 image loaded when the library is initialized, so that solid color frames
 * can be rendered. It is the default texture of various objects when no Texture object is given as
 * a parameter to its constructor.
 */
Texture* blanktexture = NULL;

/** This boolean controls whether or not the Lua UI system controls the cursor. If false, then either the
 * g3d gui subsystem or the game itself controls the mouse cursor.
 */
bool luauicursor = false;

/** The location in screen pixels of the mouse cursor. The value is only meaningful if luauicursor is
 * set to true.
 */
int luacursorx=0, luacursory=0;

extern void setg3dcxy(float x, float y);

/// cubescript interface to Lua UI library.
VAR(luauienabled, 0, 1, 1);
ICOMMAND(luauitoggle, "", (), luauienabled = luauienabled == 0 ? 1 : 0);

void luaui_grabcursor();
void luaui_releasecursor();
static int luaui_getuiparent(lua_State* L);

const char* guiEventNames[] = {
	"Hide",
	"KeyDown",
	"KeyUp",
	"MouseEnter",
	"MouseLeave",
	"MouseUp",
	"MouseDown",
	"Show",
};

enum LUAUI_BORDERSIDE
{
	LUAUI_BORDERLEFT = -1,
	LUAUI_BORDERTOP = -2,
	LUAUI_BORDERRIGHT = -3,
	LUAUI_BORDERBOTTOM = -4,
	LUAUI_CORNERTL=-5,
	LUAUI_CORNERTR=-6,
	LUAUI_CORNERBR=-7,
	LUAUI_CORNERBL=-8,
	LUAUI_DRAG=-9, // indicate drag should be started
};

/**
 * An abstract class from which all renderable objects are derived.
 *
*/
struct UIObject
{
    UIObject(UIObject* parent_, int left_, int top_, int right_, int bottom_)
        : parent(parent_), left(left_), top(top_), right(right_), bottom(bottom_),
        alpha(255), hidden(true)
    {
        if(parent == NULL)
        {
            parent = luaUIParent;
            printf("UIObject(): %p parent set to %p\n", this, parent);
        }

        if(parent != NULL)
            parent->renderlist.push_back(this);
		else 
		{
			printf(" The parent of %p is still NULL so it is the UIParent I suppose\n", this);
		}
    }

    virtual ~UIObject() {}

	/**
	 * All subclass destructors should call this method in order to detach
	 * the object from the parent's rendering list.
	*/
    void cleanup()
    {
        if(parent==NULL) 
        {
        	// printf("cleanup: %p has no parent...\n", this);
        	return;
        }

        std::vector<UIObject*>::iterator it;
        for(it = parent->renderlist.begin(); it != parent->renderlist.end(); ++it)
        {
            if((*it) == this)
            {
            	// printf("cleanup: detaching this %p from parent %p\n", this, parent);
                parent->renderlist.erase(it);
                break;
            }
        }
    }

	/**
	 * @brief Every subclass must implement this method.
	 * @param x The absolute screen position of the object's left side
	 * @param y The absolute screen position of the object's top side
	 */
    virtual void render(int x, int y) = 0;

	/**
	 * @brief A method that renders the object and all its children.
	 *
	 * This method is called by luaUIParent to render every visible UIObject.
	 */
	void renderall(int x, int y)
	{
		this->render(x,y);
		std::vector<UIObject*>::iterator it;
        for(it = renderlist.begin(); it != renderlist.end(); ++it)
        {
            if(!(*it)->hidden)
            {
                (*it)->renderall(x+(*it)->left, y+(*it)->top);
            }
        }
	}

	/**
	 * @brief Sets the alpha transparency value for an object. If allchild is true
	 * then the child objects's alpha is set to a.
	 * @param a The alpha value between 0 and 255
	 * @param allchild Set only the parent if false, else all children.
	 */
    void setalpha(int a, bool allchild)
    {
        this->alpha = a;
        // all subclass destructors are supposed to call cleanup, so this
        // iteration should be safe and not refer to any deleted pointers
        if(allchild)
        {
            for(std::vector<UIObject*>::iterator it = renderlist.begin();
                it != renderlist.end(); ++it)
            {
                (*it)->setalpha(a, allchild);
            }
        }
    }
	
	/**
	 * Sets hidden to false for the object and all its children.
	 */
	void showall()
	{
		for(std::vector<UIObject*>::iterator it = renderlist.begin();
			it != renderlist.end(); ++it)
		{
			(*it)->showall();
		}
		this->hidden = false; // parent should be shown last
	}
	
	/**
	 * Returns the absolute screen position of the object.
	 */
	void getabsxy(int &x, int &y)
	{
		int l = left;
		int t = top;
		UIObject* curpar = this->parent;
		while(curpar != NULL)
		{
			l += curpar->left;
			t += curpar->top;
			curpar = curpar->parent;
		}
		x = l;
		y = t;
	}
	
	/**
	 * Returns -1 if the point is in the resizable area, 1 if inside
	 * actual object, else 0.
	 */
	virtual int containspoint(int x, int y)
	{
		int l,t;
		getabsxy(l,t);
		int r = l + (right - left);
		int b = t + (bottom - top);
		// UIObject has no defined border so it only returns
		// a 1 or 0. subclasses may override this and return -1 sometimes.
		return (
			(x >= l) && (x <= r) &&
			(y >= t ) && (y <= b)
		) ? 1 : 0;
	}

	bool isvisible()
	{
		if(this->parent == NULL)
			return true;
		else if(this->parent->hidden)
			return false;
		else
			return !this->hidden && this->parent->isvisible();
	}

	/// the rectangle dimensions of an object are relative to the parent location.
    UIObject* parent;
    int left,
		top, 
		right, 
		bottom;
	/// A value between 0 and 255 that determines the amount of transparency of an object
    int alpha;
    /// Child objects are referenced in this list, but stored in Lua
    std::vector<UIObject*> renderlist;
    /// An object (and its children) are not rendered if this is set to false.
    bool hidden;
};

static int lua_UIObject_mm__ctor(lua_State* L)
{
	return 0;
}
#define UIOBJECT_CTOR {"__ctor", lua_UIObject_mm__ctor}

static int lua_UIObject_GetParent(lua_State* L)
{
	UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	if(self->parent == NULL || self->parent == luaUIParent)
	{
		lua_pushnil(L);
	}
	else
	{
		classbind::pushobject(L, self->parent, "UIObject");
	}

    return 1;
}

static int lua_UIObject_GetRect(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_pushinteger(L, obj->left);
    lua_pushinteger(L, obj->top);
    lua_pushinteger(L, obj->right);
    lua_pushinteger(L, obj->bottom);
    return 4;
}

static int lua_UIObject_GetAbsoluteRect(lua_State* L)
{
    UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int left = self->left;
    int top = self->top;
    int width = self->right - left;
    int height = self->bottom - top;
    UIObject* parent = self->parent;
    while(parent != NULL)
    {
        left += parent->left;
        top += parent->top;
        parent = parent->parent;
    }
    lua_pushinteger(L, left);
    lua_pushinteger(L, top);
    lua_pushinteger(L, left+width);
    lua_pushinteger(L, top+height);
    return 4;
}

static int lua_UIObject_SetRect(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    // the values are all checked first so that the operation is not
    // partially completed in case of an argument error
    //
    int newleft = luaL_checkinteger(L, 2);
    int newtop = luaL_checkinteger(L, 3);
    int newright = luaL_checkinteger(L, 4);
    int newbottom = luaL_checkinteger(L, 5);
    obj->left = newleft;
    obj->top = newtop;
    obj->right = newright;
    obj->bottom = newbottom;
    return 0;
}

static int lua_UIObject_GetWidth(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_pushinteger(L, obj->right - obj->left);
    return 1;
}

static int lua_UIObject_SetWidth(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int newwidth = luaL_checkinteger(L,2);
    obj->right = obj->left + newwidth;
    return 0;
}

static int lua_UIObject_GetHeight(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_pushinteger(L, obj->bottom - obj->top);
    return 1;
}

static int lua_UIObject_SetHeight(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int newheight = luaL_checkinteger(L, 2);
    obj->bottom = obj->top + newheight;
    return 0;
}

// UIObject.PointInRect(x,y) : boolean
static int lua_UIObject_PointInRect(lua_State* L)
{
    CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int x = luaL_checkinteger(L, 2);
    int y = luaL_checkinteger(L, 3);
    // get the absolute screen position
    lua_pushcfunction(L, lua_UIObject_GetAbsoluteRect);
    lua_pushvalue(L, 1);
    lua_call(L, 1, 4);
    int left = lua_tointeger(L, -4);
    int top = lua_tointeger(L, -3);
    int right = lua_tointeger(L, -2);
    int bottom = lua_tointeger(L, -1);
    lua_pushboolean(L, (
        (x >= left) && (x <= right) &&
        (y >= top ) && (y <= bottom)
    ));
    return 1;
}

static int lua_UIObject_ContainsPoint(lua_State* L)
{
	UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	int x = luaL_checkinteger(L, 2);
    int y = luaL_checkinteger(L, 3);
    lua_pushinteger(L, self->containspoint(x,y));
    return 1;
}

static int lua_UIObject_Resize(lua_State* L)
{
    UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_Number newwidth = luaL_checknumber(L, 2);
    lua_Number newheight = luaL_checknumber(L, 3);
    self->right = self->left + (int)newwidth;
    self->bottom = self->top + (int)newheight;
    return 0;
}

static int lua_UIObject_GetAlpha(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_pushinteger(L, obj->alpha);
    return 1;
}

// SetAlpha(alpha, [allchildren=true])
static int lua_UIObject_SetAlpha(lua_State* L)
{
    int nargs=lua_gettop(L);

    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int alpha = luaL_checkinteger(L, 2);
    bool all = true;
    if(nargs > 2)
        all = lua_toboolean(L, 3) != 0;
    obj->setalpha(alpha, all);
    return 0;
}

static int lua_UIObject_MoveTo(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int x = luaL_checkinteger(L, 2);
    int y = luaL_checkinteger(L, 3);
    int newright = x + (obj->right - obj->left);
    int newbottom = y + (obj->bottom - obj->top);
    obj->left = x;
    obj->top = y;
    obj->right = newright;
    obj->bottom = newbottom;
    return 0;
}

static int lua_UIObject_MoveRel(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int relx = luaL_checkinteger(L, 2);
    int rely = luaL_checkinteger(L, 3);
    int absx = obj->left + relx;
    int absy = obj->top + rely;
    lua_pushcfunction(L, lua_UIObject_MoveTo);
    lua_pushvalue(L, 1);
    lua_pushinteger(L, absx);
    lua_pushinteger(L, absy);
    lua_call(L, 3, 0);
    return 0;
}

static int lua_UIObject_Show(lua_State* L)
{
    UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    self->hidden = false;
    return 0;
}

static int lua_UIObject_ShowAll(lua_State* L)
{
	UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	self->showall();
	return 0;
}

static int lua_UIObject_Hide(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    obj->hidden = true;
    return 0;
}

static int lua_UIObject_IsHidden(lua_State* L)
{
    UIObject* obj = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    lua_pushboolean(L, obj->hidden);
    return 1;
}

// UIObject:RegisterEvent("event")
static int lua_UIObject_RegisterEvent(lua_State* L)
{
	CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	luaL_checkstring(L, 2);
	lua_pushlightuserdata(L, (void*)&keyListeners);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 2);
	lua_gettable(L, -2);
	lua_replace(L, -2);
	if(lua_isnil(L, -1))
	{
		lua_pop(L, 1);
		return luaL_error(L, "%s is not a valid event to register", lua_tostring(L, 2));
	}
	lua_pushvalue(L, 1);
	lua_pushvalue(L, 1);
	lua_settable(L, -3);
	lua_pop(L, 1);

	// if the third parameter is non-nil, it can be the event handler function
	//
	if(!lua_isnil(L, 3))
	{
		lua_pushliteral(L, "On");
		lua_pushvalue(L, 2);
		lua_concat(L, 2);
		lua_pushvalue(L, 3);
		lua_settable(L, 1);
	}

	return 0;
}

static int lua_UIObject_IsVisible(lua_State* L)
{
	UIObject* self = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	lua_pushboolean(L, self->isvisible() ? 1 : 0);
	return 1;
}

static const luaL_Reg lua_UIObject_methods[] = {
  {"GetParent", lua_UIObject_GetParent},
  {"Show", lua_UIObject_Show},
  {"ShowAll", lua_UIObject_ShowAll},
  {"Hide", lua_UIObject_Hide},
  {"IsHidden", lua_UIObject_IsHidden},
  {"IsVisible", lua_UIObject_IsVisible},
  {"GetRect", lua_UIObject_GetRect},
  {"SetRect", lua_UIObject_SetRect},
  {"GetAbsoluteRect", lua_UIObject_GetAbsoluteRect},
  {"GetWidth", lua_UIObject_GetWidth},
  {"SetWidth", lua_UIObject_SetWidth},
  {"GetHeight", lua_UIObject_GetHeight},
  {"SetHeight", lua_UIObject_SetHeight},
  {"PointInRect", lua_UIObject_PointInRect},
  {"ContainsPoint", lua_UIObject_ContainsPoint},
  {"Resize", lua_UIObject_Resize},
  {"MoveTo", lua_UIObject_MoveTo},
  {"MoveRel", lua_UIObject_MoveRel},
  {"GetAlpha", lua_UIObject_GetAlpha},
  {"SetAlpha", lua_UIObject_SetAlpha},
  {"RegisterEvent", lua_UIObject_RegisterEvent},
  {NULL, NULL}
};

static int luaopen_class_UIObject(lua_State* L)
{
    classbind::registerclass(L, "UIObject", lua_UIObject_methods, NULL);
    return 0;
}
ADDLUAUSERCLASS(UIObject);

/**
 * Simply a container class for child objects, automatically rendered
 * in the single renderall call by UIParent.
 */
struct UILayer : public UIObject
{
    UILayer(UIObject* parent_) : UIObject(parent_,0,0,0,0)
    {
		if(parent != NULL)
		{
			left=parent->left;
			top=parent->top;
			right=parent->right;
			bottom=parent->bottom;
		}
		else // this could be the UIParent
		{
			left=0;
			top=0;
			right=screen->w;
			bottom=screen->h;
		}
    }
    virtual ~UILayer() { cleanup(); }

    void render(int x, int y)
    {
    }
};

static int lua_UILayer_mm__new(lua_State* L)
{
    UIObject* parent = NULL;
    if(lua_isnil(L, 1) == 0)
        parent = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    UILayer* layer = new UILayer(parent);
    lua_pushlightuserdata(L, layer);
    return 1;
}

static int lua_UILayer_mm__delete(lua_State* L)
{
    UILayer* layer = (UILayer*)lua_touserdata(L, 1);
    delete layer;
    return 0;
}

static const luaL_Reg lua_UILayer_methods[] =
{
	{NULL, NULL}
};

static const luaL_Reg lua_UILayer_metamethods[] =
{
	{"__new", lua_UILayer_mm__new},
	{"__delete", lua_UILayer_mm__delete},
	UIOBJECT_CTOR,
	{NULL, NULL}
};

static int luaopen_class_UILayer(lua_State* L)
{
	classbind::registerclass(L, "UILayer", lua_UILayer_methods, lua_UILayer_metamethods);
	CLASSBIND_ADDPARENT(L, UILayer, UIObject);
    return 0;
}
ADDLUAUSERCLASS(UILayer);

/**
 * This is a type of renderable object that supports simple borders, rendering an
 * image, or being a solid color.
 */
struct UIFrame : public UIObject
{
	/**
	 * This is used internally by a %UIFrame to hold border information for rendering.
	 * It cannot be directly accessed from Lua but is created when border set methods
	 * are called.
	 */
    struct Border
    {
        Texture* texture;
		int width;
        int x, y;
        int r,g,b;
        float cornertw; // defaults to 0.5
        bool nocorner;
        int inset;
    };

    UIFrame(UIObject* parent_, int width, int height, Texture* texture_, int r, int g, int b)
        : UIObject(parent_, 0,0,width,height),
        red(r), green(g), blue(b),
        texture(texture_),
        texleft(0),textop(0),texright(1),texbottom(1),
        border(NULL)
    {
        if(texture == NULL)
            texture = blanktexture;
    }

    virtual ~UIFrame() { delete border; cleanup(); }

    void render(int x, int y)
    {
		int x2 = x + right - left;
		int y2 = y + bottom - top;
        glColor4ub(this->red, this->green, this->blue, this->alpha);
        glBindTexture(GL_TEXTURE_2D, this->texture->id);
        glBegin(GL_QUADS);
        glTexCoord2f(texleft,textop);       glVertex2f(x,y);
        glTexCoord2f(texright,textop);      glVertex2f(x2,y);
        glTexCoord2f(texright,texbottom);   glVertex2f(x2,y2);
        glTexCoord2f(texleft,texbottom);    glVertex2f(x,y2);
        glEnd();
		// if there is a border to render, render it
		if(border)
		{
			glColor4ub(border->r, border->g, border->b, this->alpha);
			if(border->inset != 0)
			{
				x += border->inset;
				x2 -= border->inset;
				y += border->inset;
				y2 -= border->inset;
			}
			int bleft = x - border->width;
			int btop = y - border->width;
			int bright = x2 + border->width;
			int bbottom = y2 + border->width;
			float cx = border->nocorner ? 0 : border->cornertw;
			if(border->texture)
				glBindTexture(GL_TEXTURE_2D, border->texture->id);
			else 
				glBindTexture(GL_TEXTURE_2D, blanktexture->id);
			// render y
			glBegin(GL_QUADS);
			glTexCoord2f(cx,0);		glVertex2f(x, btop);
			glTexCoord2f(1,0);		glVertex2f(x2, btop);
			glTexCoord2f(1,1);		glVertex2f(x2, y);
			glTexCoord2f(cx,1);		glVertex2f(x, y);
			glEnd();
			// render x2 side
			glBegin(GL_QUADS);
			glTexCoord2f(cx,1);		glVertex2f(x2, y);
			glTexCoord2f(cx,0);		glVertex2f(bright, y);
			glTexCoord2f(1,0); 		glVertex2f(bright, y2);
			glTexCoord2f(1,1); 		glVertex2f(x2, y2);
			glEnd();
			// render y2 side
			glBegin(GL_QUADS);
			glTexCoord2f(1,1);		glVertex2f(x, y2);
			glTexCoord2f(cx,1);		glVertex2f(x2, y2);
			glTexCoord2f(cx,0);		glVertex2f(x2, bbottom);
			glTexCoord2f(1,0);		glVertex2f(x, bbottom);
			glEnd();
			// render x side
			glBegin(GL_QUADS);
			glTexCoord2f(1,0);		glVertex2f(bleft, y);
			glTexCoord2f(1,1);		glVertex2f(x, y);
			glTexCoord2f(cx,1);		glVertex2f(x, y2);
			glTexCoord2f(cx,0);		glVertex2f(bleft, y2);
			glEnd();
			if(border->nocorner == false)
			{
				// y x corner
				glBegin(GL_QUADS);
				glTexCoord2f(0,0);	glVertex2f(bleft, btop);
				glTexCoord2f(cx,0); glVertex2f(x, btop);
				glTexCoord2f(cx,1); glVertex2f(x, y);
				glTexCoord2f(0,1);  glVertex2f(bleft, y);
				glEnd();
				// y x2 corner
				glBegin(GL_QUADS);
				glTexCoord2f(0,1);	glVertex2f(x2, btop);
				glTexCoord2f(0,0);	glVertex2f(bright, btop);
				glTexCoord2f(cx,0); glVertex2f(bright, y);
				glTexCoord2f(cx,1); glVertex2f(x2, y);
				glEnd();
				// y2 x2 corner
				glBegin(GL_QUADS);
				glTexCoord2f(cx,1);	glVertex2f(x2, y2);
				glTexCoord2f(0,1);	glVertex2f(bright, y2);
				glTexCoord2f(0,0);  glVertex2f(bright, bbottom);
				glTexCoord2f(cx,0); glVertex2f(x2, bbottom);
				glEnd();
				// y2 x corner
				glBegin(GL_QUADS);
				glTexCoord2f(cx,0);	glVertex2f(bleft, y2);
				glTexCoord2f(cx,1);	glVertex2f(x, y2);
				glTexCoord2f(0,1);	glVertex2f(x, bbottom);
				glTexCoord2f(0,0);	glVertex2f(bleft, bbottom);
				glEnd();
			}
		}
    }

    int red, green, blue;
    Texture* texture;
    float texleft, textop, texright,texbottom;
    Border* border;
};

static int lua_UIFrame_mm__new(lua_State* L)
{
	int numArgs = lua_gettop(L);
	printf("UIFrame: numArgs=%d\n", numArgs);
    UIObject* parent = NULL;
    if(lua_isnil(L, 1)==0)
        parent = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    int width = luaL_checkinteger(L, 2);
    int height = luaL_checkinteger(L, 3);
    Texture* texture = blanktexture;
    if(numArgs>=4)
	{
		printf("UIFrame: There is indeed a specified Texture,");
		if(lua_isnil(L, 4)==0)
		{
			texture = CLASSBIND_CHECKOBJECT(L, 4, Texture);
			printf(" 0x%x", texture);
		}
		printf("\n");
	}
    int red = luaL_optinteger(L, 5, 255);
    int green = luaL_optinteger(L, 6, 255);
    int blue = luaL_optinteger(L, 7, 255);

    UIFrame* frame = new UIFrame(parent, width, height, texture, red, green, blue);

    lua_pushlightuserdata(L, frame);

    return 1;
}

static int lua_UIFrame_mm__delete(lua_State* L)
{
    UIFrame* frame = (UIFrame*)lua_touserdata(L, 1);
    delete frame;
    return 0;
}

// UIFrame:SetBorder
static int lua_UIFrame_SetBorder(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
	if(self->border)
	{
		delete self->border;
		self->border = NULL;
	}
	Texture* btx = NULL;
	if(!lua_isnil(L, 2))
		btx = CLASSBIND_CHECKOBJECT(L, 2, Texture);
	int width = luaL_optinteger(L, 3, 1);
	int inset = luaL_optinteger(L, 4, 0);
	bool nocorner = lua_toboolean(L, 5) != 0;
	float tw = luaL_optnumber(L, 6, 0.5f);
	int r = luaL_optinteger(L, 7, 255);
	int g = luaL_optinteger(L, 8, 255);
	int b = luaL_optinteger(L, 9, 255);
	
	self->border = new UIFrame::Border();
	if(!self->border)
	{
		return luaL_error(L, "Failed to create new Border");
	}
	self->border->texture = btx;
	self->border->width = width;
	self->border->r = r;
	self->border->g = g;
	self->border->b = b;
	self->border->cornertw = tw;
	self->border->nocorner = nocorner;
	self->border->inset = inset;
	return 0;
}

// change existing border's color or make a new solid color one
static int lua_UIFrame_SetBorderColor(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
	int r = luaL_checkinteger(L, 2);
	int g = luaL_checkinteger(L, 3);
	int b = luaL_checkinteger(L, 4);
	int width = luaL_optinteger(L, 5, 1);
	if(self->border)
	{
		self->border->r = r;
		self->border->g = g;
		self->border->b = b;
	}
	else 
	{
		self->border = new UIFrame::Border();
		if(!self->border) return luaL_error(L, "Failed to create Border");
		self->border->texture = NULL;
		self->border->inset = 0;
		self->border->width = width;
		self->border->nocorner = false;
		self->border->cornertw = 0.5f;
		self->border->r = r;
		self->border->g = g;
		self->border->b = b;
	}
	return 0;
}

static int lua_UIFrame_SetColor(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
    int red = luaL_checkinteger(L, 2);
    int green = luaL_checkinteger(L, 3);
    int blue = luaL_checkinteger(L, 4);
    self->red = red;
    self->blue = blue;
    self->green = green;
    return 0;
}

static int lua_UIFrame_SetTexture(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
	if(!lua_isnil(L, 2))
	{
		Texture* tx = CLASSBIND_CHECKOBJECT(L, 2,Texture);
		self->texture = tx;
	}
	else
		self->texture = blanktexture;
	return 0;
}

static int lua_UIFrame_GetTexture(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
	classbind::pushobject(L, self->texture, "Texture");
	return 1;
}

static int lua_UIFrame_SetTexCoords(lua_State* L)
{
	UIFrame* self = CLASSBIND_CHECKOBJECT(L, 1, UIFrame);
	lua_Number texleft = luaL_checknumber(L, 2);
	lua_Number textop = luaL_checknumber(L, 3);
	lua_Number texright = luaL_checknumber(L, 4);
	lua_Number texbottom = luaL_checknumber(L, 5);
	self->texleft = texleft;
	self->textop = textop;
	self->texright = texright;
	self->texbottom = texbottom;
	return 0;
}

static const luaL_Reg lua_UIFrame_methods[] =
{
    {"SetColor", lua_UIFrame_SetColor},
    {"SetBorder", lua_UIFrame_SetBorder},
    {"SetBorderColor", lua_UIFrame_SetBorderColor},
    {"SetTexture", lua_UIFrame_SetTexture},
    {"GetTexture", lua_UIFrame_GetTexture},
    {"SetTexCoords", lua_UIFrame_SetTexCoords},
    {NULL, NULL}
};

static const luaL_Reg lua_UIFrame_metamethods[] =
{
	{"__new", lua_UIFrame_mm__new},
    {"__delete", lua_UIFrame_mm__delete},
    UIOBJECT_CTOR,
    {NULL, NULL}
};

static int luaopen_class_UIFrame(lua_State* L)
{
    classbind::registerclass(L, "UIFrame", lua_UIFrame_methods, lua_UIFrame_metamethods);
	CLASSBIND_ADDPARENT(L, UIFrame, UIObject);
    return 0;
}
ADDLUAUSERCLASS(UIFrame);

/**
 * This holds unformatted text with optional color and scaling.
 */
struct UIText : UIObject
{
    UIText(UIObject* parent_, const char* text_, float scale_, int r, int g, int b)
        : UIObject(parent_, 0,0,0,0), text(text_), scale(scale_),
        red(r), green(g), blue(b)
    {
        // set default size according to text width
        //
        int _w, _h;
        text_bounds(text.c_str(), _w, _h, -1);
        right = (int)((float)_w * scale);
        bottom = (int)((float)_h * scale);
    }

    virtual ~UIText() { cleanup(); }

    void render(int x, int y)
    {
		// we have to push the matrix and translate due to text scaling
		//
        glPushMatrix();
        glTranslatef(x, y, 1.0f);
        glScalef(scale,scale,0.0f);
        draw_text(text.c_str(), 0,0,red,green,blue,alpha,-1,-1);
        glPopMatrix();
    }

    std::string text;
    float scale;
    int red, green, blue;
};

// UIText:FitText(wbound, hbound)
static int lua_UIText_FitText(lua_State* L)
{
    // try to scale text down so that it fits inside bounds
    //
	UIText* self = CLASSBIND_CHECKOBJECT(L, 1, UIText);
    int wbound = luaL_checkinteger(L,2);
    int hbound = luaL_checkinteger(L,3);

    int _width = 0;
    int _height = 0;
    text_bounds(self->text.c_str(), _width, _height, -1);

    float height = (float)_height * self->scale;

    if(height > hbound)
    {
        self->scale = self->scale * static_cast<float>(hbound) / static_cast<float>(height);
        // printf("Changed scale to %f\n", label->scale);
    }

    float width = (float)_width * self->scale;

    if(width > wbound)
    {
        self->scale = self->scale * static_cast<float>(wbound) / static_cast<float>(width);
        // printf(" Changed scale to %f\n", label->scale);
    }

    return 0;
}

// UIText.SetText(txt, [r,g,b])
static int lua_UIText_SetText(lua_State* L)
{
	UIText* self = CLASSBIND_CHECKOBJECT(L, 1, UIText);
	self->text = luaL_checkstring(L, 2);
	self->red = luaL_optinteger(L, 3, self->red);
	self->green = luaL_optinteger(L, 4, self->green);
	self->blue = luaL_optinteger(L, 5, self->blue);
	return 0;
}

// UIText.GetText() : string, r,g,b
static int lua_UIText_GetText(lua_State* L)
{
	UIText* self = CLASSBIND_CHECKOBJECT(L, 1, UIText);
	lua_pushstring(L, self->text.c_str());
	lua_pushinteger(L, self->red);
	lua_pushinteger(L, self->green);
	lua_pushinteger(L, self->blue);
	return 4;
}

static int lua_UIText_Resize(lua_State* L)
{
	UIText* self = CLASSBIND_CHECKOBJECT(L, 1, UIText);
    lua_pushcfunction(L, lua_UIObject_Resize);
    lua_pushvalue(L, 1);
    lua_pushvalue(L, 2);
    lua_pushvalue(L, 3);
    lua_call(L, 3, 0);
    self->scale = self->bottom - self->top; // a font's height should at least be 1 pixel
    lua_pushcfunction(L, lua_UIText_FitText);
    lua_pushvalue(L, 1);
    lua_pushinteger(L, self->right - self->left);
    lua_pushinteger(L, self->bottom - self->top);
    lua_call(L, 3, 0);
    return 0;
}

static int lua_UIText_mm__new(lua_State* L)
{
    UIObject* parent = NULL;
    if(!lua_isnil(L,1))
        parent = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
    const char* text = luaL_checkstring(L, 2);
    float scale = luaL_optnumber(L, 3, 0.5f);
    int red = luaL_optinteger(L, 4, 255);
    int green = luaL_optinteger(L, 5, 255);
    int blue = luaL_optinteger(L, 6, 255);
    UIText* label = new UIText(parent,text,scale,red,green,blue);
    lua_pushlightuserdata(L, label);
    return 1;
}

static int lua_UIText_mm__delete(lua_State* L)
{
    UIText* label = (UIText*)lua_touserdata(L, 1);
    delete label;
    return 0;
}

static const luaL_Reg lua_UIText_methods[] =
{
    {"FitText", lua_UIText_FitText},
	{"SetText", lua_UIText_SetText},
	{"GetText", lua_UIText_GetText},
    {"Resize", lua_UIText_Resize},
    {NULL, NULL}
};

static const luaL_Reg lua_UIText_metamethods[] =
{
	{"__new", lua_UIText_mm__new},
    {"__delete", lua_UIText_mm__delete},
    UIOBJECT_CTOR,
    {NULL, NULL}
};

static int luaopen_class_UIText(lua_State* L)
{
    classbind::registerclass(L,"UIText", lua_UIText_methods, lua_UIText_metamethods);
	CLASSBIND_ADDPARENT(L, UIText, UIObject);
    return 0;
}
ADDLUAUSERCLASS(UIText);

/**
 * @class UIDialog
 * A class that creates a dialog that looks similar to the old 3d GUI dialogs by
 * using guiskin.png and chopping it up similarly.
 */
static const int s_guiSkinX[] = {0, 11, 23, 37, 105, 119, 137, 151, 215, 229, 246, 256};
static const int s_guiSkinY[] = {0, 7, 21, 34, 43, 48, 56, 104, 111, 117, 128};
static const struct GuiPatch_ {

	// indices into guiSkin[X/Y] array
	int left, right, top, bottom;

} s_guiPatches[] = {
	{1,2,3,6}, // 0
	{2,9,5,6}, // content frame
	{9,10,3,6}, // 2
	{1,2,6,7}, // 3
	{2,9,6,7}, // 4
	{9,10,6,7}, // 5

	{1,2,7,9}, // 6
	{2,9,7,9}, // 7
	{9,10,7,9}, // 8

	{5,6,3,5}, // 9

	{2,3,1,2}, // 10
	{3,4,1,2}, // 11
	{4,5,1,2}, // 12
	{2,3,2,3}, // 13
	{3,4,2,3}, // 14
	{4,5,2,3}, // 15
	{2,3,3,5}, // 16
	{3,4,3,5}, // 17
	{4,5,3,5}, // 18

	{6,7,1,2},
	{7,8,1,2},
	{8,9,1,2},
	{6,7,2,3},
	{7,8,2,3},
	{8,9,2,3},
	{6,7,3,5},
	{7,8,3,5},
	{8,9,3,5},
};

struct UIDialog : public UIObject 
{
	static int s_topPaddingHeight;
	static Texture* s_guiTexture;
	static int s_maxTexW;
	static int s_maxTexH;
	
	std::string title;
	bool resizable;
	
	UIDialog(UIObject* parent_, int width, int height)
		: UIObject(parent_, 0, 0, width, height)
	{
		if(s_guiTexture == NULL)
			s_guiTexture = textureload("data/guiskin.png");
			
		if(s_maxTexW == 0 || s_maxTexH == 0)
		{
			float l,t,r,b;
			getTexCoords(l,t,r,b, s_guiPatches[4]);
			s_maxTexW = (r - l) * (float)s_guiTexture->w;
			s_maxTexH = (b - t) * (float)s_guiTexture->h;
			if(s_maxTexW == 0) s_maxTexW = 1;
			if(s_maxTexH == 0) s_maxTexH = 1;
		}
		
		if(s_topPaddingHeight == 0)
		{
			static const GuiPatch_ special={2,3,3,4};
			int _w;
			getPatchSize(_w, s_topPaddingHeight, special);
			s_topPaddingHeight -= 2;
		}
		
		// conoutf("%d, %d", UIDialog_maxTexW, UIDialog_maxTexH);
	}
	
	virtual ~UIDialog() { cleanup(); }
	
	static void getborderwidth(int &left, int &top, int &right, int &bottom)
	{
		
	}
	
	virtual int containspoint(int x, int y)
	{
		if( UIObject::containspoint(x,y) == 0 )
		{
			// is it on our borders?
			int l,t,r,b;
			getabsxy(l,t);
			r = l + (right - left);
			b = t + (bottom - top);
			// todo: finish this
			return 0;
		}
		else
		{
			return 1;
		}
	}
	
	static void renderPartRect(int x, int y, int w, int h, float tl, float tt,
								float tr, float tb)
	{
		// width and height defaults to size of the texture patch
		//
		int patchWidth = (tr - tl) * (float)(s_guiTexture->w);
		int patchHeight = (tb - tt) * (float)(s_guiTexture->h);
		if (w < 0) w = patchWidth;
		if (h < 0) h = patchHeight;
		
		int numPartsY = h / patchHeight;
		int numPartsX = w / patchWidth;
		int partHeight = h % patchHeight;
		int partWidth = w % patchWidth;
		
		float tr_ = tl + (float)partWidth / (float)s_guiTexture->w, 
			tb_ = tt + (float)partHeight / (float)s_guiTexture->h;
		
		glBegin(GL_QUADS);
		// render all full pieces first
		int x_ = x;
		for(int iy = 0; iy < numPartsY; ++iy)
		{
			for(int ix = 0; ix < numPartsX; ++ix)
			{
				glTexCoord2f(tl,tt);	glVertex2f(x,y);
				glTexCoord2f(tr,tt);	glVertex2f(x+patchWidth,y);
				glTexCoord2f(tr,tb);	glVertex2f(x+patchWidth,y+patchHeight);
				glTexCoord2f(tl,tb);	glVertex2f(x,y+patchHeight);
				x += patchWidth;
			}
			if(partWidth > 0)
			{
				glTexCoord2f(tl,tt);	glVertex2f(x,y);
				glTexCoord2f(tr_,tt);	glVertex2f(x+partWidth,y);
				glTexCoord2f(tr_,tb);	glVertex2f(x+partWidth,y+patchHeight);
				glTexCoord2f(tl,tb);	glVertex2f(x,y+patchHeight);
			}
			y += patchHeight;
			
			x = x_;
		}
		if(partHeight > 0)
		{
			for(int ix = 0; ix < numPartsX; ++ix)
			{
				glTexCoord2f(tl,tt);	glVertex2f(x,y);
				glTexCoord2f(tr,tt);	glVertex2f(x+patchWidth,y);
				glTexCoord2f(tr,tb_);	glVertex2f(x+patchWidth,y+partHeight);
				glTexCoord2f(tl,tb_);	glVertex2f(x,y+partHeight);
				x += patchWidth;
			}
			if(partWidth > 0)
			{
				glTexCoord2f(tl,tt);	glVertex2f(x,y);
				glTexCoord2f(tr_,tt);	glVertex2f(x+partWidth,y);
				glTexCoord2f(tr_,tb_);	glVertex2f(x+partWidth,y+partHeight);
				glTexCoord2f(tl,tb_);	glVertex2f(x,y+partHeight);
			}
			x = x_;
		}
		glEnd();
	}
	
	static void getPatchSize(int &w, int &h, const GuiPatch_ & patch)
	{
		float tl, tr, tt, tb;
		getTexCoords(tl,tt,tr,tb,patch);
		w = (tr-tl) * (float)s_guiTexture->w;
		h = (tb - tt) * (float)s_guiTexture->h;
	}
	
	static void getTexCoords(float &tleft, float &ttop, float &tright, float &tbottom, 
							const GuiPatch_ & patch)
	{
		tleft = (float)s_guiSkinX[patch.left] / 256.0f;
		tright = (float)s_guiSkinX[patch.right] / 256.0f;
		ttop = (float)s_guiSkinY[patch.top] / 128.0f;
		tbottom = (float)s_guiSkinY[patch.bottom] / 128.0f;
	}
	
	void renderTitleBar(int x, int y)
	{
		float tl, tt, tr, tb;
		int w,h;
		// lowest left corner
		getPatchSize(w, h, s_guiPatches[16]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[16]);
		renderPartRect(x, y - h - s_topPaddingHeight,  -1, -1, tl, tt, tr, tb);
		int h1 = h; // save this height for later
		// bar across lowest
		getTexCoords(tl, tt, tr, tb, s_guiPatches[17]);
		renderPartRect(x + w, y - h1 - s_topPaddingHeight, (right - left) - (2 * w), -1,
						tl, tt, tr, tb);
		// lowest right corner
		getPatchSize(w, h, s_guiPatches[18]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[18]);
		renderPartRect(x + (right - left) - w, y - h1 - s_topPaddingHeight, -1, -1, tl, tt, tr, tb);
		y -= h1 + s_topPaddingHeight;
		// next left corner
		getPatchSize(w, h, s_guiPatches[13]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[13]);
		renderPartRect(x, y - h, -1, -1, tl, tt, tr, tb);
		// next bar across
		getTexCoords(tl, tt, tr, tb, s_guiPatches[14]);
		renderPartRect(x + w, y - h, (right - left) - (2 * w), -1,
						tl, tt, tr, tb);
		// next right corner
		getPatchSize(w, h, s_guiPatches[15]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[15]);
		renderPartRect(x + (right - left) - w, y - h, -1, -1, tl, tt, tr, tb);
		y -= h;
		// highest top left corner
		// glColor4ub(224, 216, 216, alpha);
		getPatchSize(w, h, s_guiPatches[10]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[10]);
		renderPartRect(x, y - h, -1, -1, tl, tt, tr, tb);
		// highest bar across
		getTexCoords(tl,tt,tr,tb, s_guiPatches[11]);
		renderPartRect(x + w, y - h, (right - left) - (2*w), -1, tl, tt, tr, tb);
		// highest right corner
		getPatchSize(w, h, s_guiPatches[12]);
		getTexCoords(tl, tt, tr, tb, s_guiPatches[12]);
		renderPartRect(x + (right - left) - w, y - h, -1, -1, tl, tt, tr, tb);
		
		glPushMatrix();
        glTranslatef(x + w, y - h / 2, 1.0f);
        glScalef(0.4,0.4, 1.f);
        int textwidth, textheight;
        std::string renderText = title;
        int maxlen = renderText.length();
        text_bounds((char*)title.c_str(), textwidth, textheight, -1);
        while(textwidth*0.4 > (right - left) - 2* w && maxlen > 1)
        {
			text_bounds((char*)renderText.substr(0,maxlen).c_str(), textwidth, textheight, -1);
			maxlen--;
        }
        renderText = renderText.substr(0,maxlen);
        draw_text(renderText.c_str(), 0,0, 255, 255, 240, 255, -1,-1);
        glPopMatrix();
	}
	
	void render(int x, int y)
	{
		if (right - left < 60)
			right = left + 60;
		glColor4ub(255, 255, 255, alpha);
		glBindTexture(GL_TEXTURE_2D, s_guiTexture->id);
		float tleft, ttop, tright, tbottom;
		int w,h;
		// the body
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[4]);
		renderPartRect(x, y, right - left, bottom - top, tleft, ttop, tright, tbottom);
		renderPartRect(x, y - s_topPaddingHeight, right - left, s_topPaddingHeight, 
						tleft, ttop, tright, tbottom);
		// the top left corner
		getPatchSize(w,h,s_guiPatches[0]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[0]);
		renderPartRect(x - w, y - h, -1, -1, tleft, ttop, tright, tbottom);
		// the left side
		getPatchSize(w, h, s_guiPatches[3]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[3]);
		renderPartRect(x - w, y, -1, (bottom - top), tleft, ttop, tright, tbottom);
		// the bottom left corner
		getPatchSize(w,h,s_guiPatches[6]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[6]);
		renderPartRect(x - w, y+(bottom-top), -1, -1, tleft, ttop, tright, tbottom);
		// the bottom side
		getPatchSize(w, h, s_guiPatches[7]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[7]);
		renderPartRect(x, y+(bottom-top), (right - left), -1, tleft, ttop, tright, tbottom);
		// bottom right corner
		getPatchSize(w, h, s_guiPatches[8]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[8]);
		renderPartRect(x + (right - left), y + (bottom-top), -1,-1, tleft,ttop,tright,tbottom);
		// right side
		getPatchSize(w, h, s_guiPatches[5]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[5]);
		renderPartRect(x + (right-left), y, -1, (bottom-top), tleft,ttop,tright,tbottom);
		// top right corner
		getPatchSize(w, h, s_guiPatches[2]);
		getTexCoords(tleft, ttop, tright, tbottom, s_guiPatches[2]);
		renderPartRect(x + (right-left), y - h, -1,-1, tleft,ttop,tright,tbottom);
		// top
		renderTitleBar(x,y);
	}
};
int UIDialog::s_topPaddingHeight = 0;
Texture* UIDialog::s_guiTexture = NULL;
int UIDialog::s_maxTexW = 0;
int UIDialog::s_maxTexH = 0;

static void setDlgStatus(lua_State* L)
{
	bool grab = false;
	
	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	lua_gettable(L, LUA_REGISTRYINDEX);
	
	// grab = lua_objlen(L, -1) != 0;
	grab = getNumList(L, -1) != 0;
	printf("Shown list size: %d\n", getNumList(L, -1));
	lua_pop(L, 1);
	
	if(grab) luaui_grabcursor();
	else luaui_releasecursor();
}

static int lua_UIDialog_mm__new(lua_State* L)
{
	UIObject *parent = NULL;
	if(!lua_isnil(L, 1))
		parent = CLASSBIND_CHECKOBJECT(L,1,UIObject);
	int width = luaL_checkinteger(L, 2);
	int height = luaL_checkinteger(L, 3);
	UIDialog* dialog = new UIDialog(parent, width, height);
	lua_pushlightuserdata(L, dialog);
	return 1;
}

static int lua_UIDialog_mm__delete(lua_State* L)
{
	UIDialog* dlg = (UIDialog*)lua_touserdata(L, 1);
	delete dlg;
	return 0;
}

static int lua_UIDialog_mm__dtor(lua_State* L)
{
	// remove from shown list
	// (this shouldn't be needed since it is a weak reference table, but it
	// seems to be undefined whether or not a weak reference exists during the
	// call to __gc).
	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_pushnil(L);
	lua_settable(L, -3);
	lua_pop(L, 1);
	setDlgStatus(L);
	return 0;
}

static int lua_UIDialog_SetTitle(lua_State* L)
{
	UIDialog* self = CLASSBIND_CHECKOBJECT(L,1,UIDialog);
	self->title = std::string( luaL_checkstring(L, 2) );
	return 0;
}

static int lua_UIDialog_GetTitle(lua_State* L)
{
	UIDialog* self = CLASSBIND_CHECKOBJECT(L,1,UIDialog);
	lua_pushstring(L, self->title.c_str());
	return 1;
}

static int lua_UIDialog_Show(lua_State* L)
{
	// super.Show()
	lua_pushcfunction(L, lua_UIObject_Show);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 0);
	// add to shown list
	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_pushvalue(L, 1);
	lua_settable(L, -3);
	lua_pop(L, 1);
	setDlgStatus(L);
	return 0;
}

static int lua_UIDialog_ShowAll(lua_State* L)
{
	// super.ShowAll()
	lua_pushcfunction(L, lua_UIObject_ShowAll);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 0);
	// add to shown list
	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_pushvalue(L, 1);
	lua_settable(L, -3);
	lua_pop(L, 1);
	setDlgStatus(L);
	return 0;	
}

static int lua_UIDialog_Hide(lua_State* L)
{
	// super.Hide()
	lua_pushcfunction(L, lua_UIObject_Hide);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 0);
	// remove from shown list
	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_pushnil(L);
	lua_settable(L, -3);
	lua_pop(L, 1);
	setDlgStatus(L);
	return 0;
}

static const luaL_Reg lua_UIDialog_methods[] = {
	{"Show", lua_UIDialog_Show},
	{"ShowAll", lua_UIDialog_ShowAll},
	{"Hide", lua_UIDialog_Hide},
	{"GetTitle", lua_UIDialog_GetTitle},
	{"SetTitle", lua_UIDialog_SetTitle},
	{NULL, NULL}
};

static const luaL_Reg lua_UIDialog_metamethods[] = {
	{"__new", lua_UIDialog_mm__new},
	{"__delete", lua_UIDialog_mm__delete},
	{"__dtor", lua_UIDialog_mm__dtor},
	UIOBJECT_CTOR,
	{NULL, NULL}
};

static int luaopen_class_UIDialog(lua_State* L)
{
	classbind::registerclass(L, "UIDialog", lua_UIDialog_methods, lua_UIDialog_metamethods);
	CLASSBIND_ADDPARENT(L, UIDialog, UIObject);
	return 0;
}
ADDLUAUSERCLASS(UIDialog);

/**
Button that is drawn in the style of UIDialog's texture.
*/
struct UIButton : public UIObject
{
	UIButton(UIObject* parent_, int x, int y, int width, int height)
		: UIObject(parent_, x, y, x+width, y+height), labelImage(NULL), labelText("")
	{
		if(UIDialog::s_guiTexture == NULL)
			UIDialog::s_guiTexture = textureload("data/guiskin.png");
	}

	virtual ~UIButton() { cleanup(); }

	void render(int x, int y)
	{
		static Texture* btnTexture = textureload("data/guioverlay.png");
		if(!btnTexture) return;
		
		int x2 = x + right - left;
		int y2 = y + bottom - top;
		if(labelImage == NULL)
		{
			int _width = 0;
			int _height = 0;
			float scale=1.0f;
			text_bounds(labelText.c_str(), _width, _height, -1);
			if(_height > bottom - top)
			{
				scale = static_cast<float>(bottom - top) / static_cast<float>(_height);
			}
			if(_width * scale > right - left)
			{
				scale = static_cast<float>(right - left) / static_cast<float>(_width);
			}
			// scale should be 0.4 or smaller
			if(scale > 0.4f) { scale = 0.4f; }
			// printf("UIButton::render: Label length=%d\n", _width);
			// printf("UIButton::render: Button length=%d\n", right - left);
			// printf("UIButton::render: Label scale is now %f\n", scale);
			glPushMatrix();
				glTranslatef(x, y, 1.0f);
				glScalef(scale,scale,0.0f);
				draw_text(labelText.c_str(), 0,0,255,255,255,alpha,-1,-1);
			glPopMatrix();
		}
		glBindTexture(GL_TEXTURE_2D, btnTexture->id);
		glBegin(GL_QUADS);
			// glTexCoord2f(tl,tt);	glVertex2f(x,y);
			glTexCoord2f(0.0f, 0.0f);	glVertex2f(x,y);
			glTexCoord2f(1.0f, 0.0f);	glVertex2f(x2, y);
			glTexCoord2f(1.0f, 1.0f);	glVertex2f(x2, y2);
			glTexCoord2f(0.0f, 1.0f);	glVertex2f(x, y2);
		glEnd();
	}

	std::string labelText;
	Texture* labelImage;

	bool pressed;
};

static int lua_UIButton_mm__new(lua_State* L)
{
	int numArgs = lua_gettop(L);
	UIObject* parent = NULL;
	if(!lua_isnil(L, 1))
		parent = CLASSBIND_CHECKOBJECT(L, 1, UIObject);
	int x = luaL_checkinteger(L, 2);
	int y = luaL_checkinteger(L, 3);
	int width = luaL_checkinteger(L, 4);
	int height = luaL_checkinteger(L, 5);
	Texture* imgLabel = NULL;
	std::string txtLabel = "<UIButton>";
	if(numArgs >= 6)
	{
		if(classbind::isobject(L, 6, "Texture"))
			imgLabel =  CLASSBIND_CHECKOBJECT(L, 6, Texture);
		else
			txtLabel = std::string( luaL_checkstring(L, 6) );
	}
	UIButton* button = new UIButton(parent, x, y, width, height);
	button->labelImage = imgLabel;
	if(!button->labelImage) // either a text-only label or an image label, not both
		button->labelText = txtLabel;

	lua_pushlightuserdata(L, button);
	return 1;
}

static int lua_UIButton_mm__ctor(lua_State* L)
{
	// add all UIButton objects to list
	lua_pushlightuserdata(L, (void*)&keyButtons);
	lua_gettable(L, LUA_REGISTRYINDEX);
	lua_pushvalue(L, 1);
	lua_pushvalue(L, 1);
	lua_settable(L, -3);
	lua_pop(L, 1);
	return 0;
}

static int lua_UIButton_mm__delete(lua_State* L)
{
	UIButton* button = (UIButton*)lua_touserdata(L, 1);
	delete button;
	return 0;
}

static const luaL_Reg lua_UIButton_methods[] = {
	{NULL, NULL}
};

static const luaL_Reg lua_UIButton_metamethods[] = {
	{"__new", lua_UIButton_mm__new},
	{"__ctor", lua_UIButton_mm__ctor},
	{"__delete", lua_UIButton_mm__delete},
	{NULL, NULL}
};

static int luaopen_class_UIButton(lua_State* L)
{
	classbind::registerclass(L, "UIButton", lua_UIButton_methods, lua_UIButton_metamethods);
	CLASSBIND_ADDPARENT(L, UIButton, UIObject);
	return 0;
}
ADDLUAUSERCLASS(UIButton);

// free functions defined by luaui

int lua_HasCursor(lua_State* L)
{
	lua_pushboolean(L, luauicursor);
	return 1;
}
ADDLUAUSERAPI(HasCursor);

int lua_GrabCursor(lua_State* L)
{
	// luauicursor = true;
	luaui_grabcursor();
	return 0;
}
ADDLUAUSERAPI(GrabCursor);

int lua_ReleaseCursor(lua_State* L)
{
	// luauicursor = false;
	luaui_releasecursor();
	return 0;
}
ADDLUAUSERAPI(ReleaseCursor);

// textbounds
int lua_GetTextBounds(lua_State* L)
{
	const char* text = luaL_checkstring(L,1);
	int w,h;
	text_bounds((char*)text, w, h, -1);
	lua_pushinteger(L, w);
	lua_pushinteger(L, h);
	return 2;
}
ADDLUAUSERAPI(GetTextBounds);

int lua_GetCursorX(lua_State* L)
{
	lua_pushinteger(L, luacursorx);
	return 1;
}
ADDLUAUSERAPI(GetCursorX);

int lua_GetCursorY(lua_State* L)
{
	lua_pushinteger(L, luacursory);
	return 1;
}
ADDLUAUSERAPI(GetCursorY);

int lua_GetCursorPosition(lua_State* L)
{
	lua_pushinteger(L, luacursorx);
	lua_pushinteger(L, luacursory);
	return 2;
}
ADDLUAUSERAPI(GetCursorPosition);

// game interface

void luaui_grabcursor()
{
	luauicursor = true;
	luacursorx = screen->w / 2;
	luacursory = screen->h / 2;
}

void luaui_releasecursor()
{
	luauicursor = false;
	setg3dcxy(0.5,0.5);
}

/**
 * This is called by the game engine each time the mouse position is changed.
 * It sets the values of luacursorx and luacursory to the new mouse position
 * and returns true if Lua UI is holding the cursor.
 */
bool luaui_movecursor(int x, int y)
{
	if(!luauicursor) return false;
	luacursorx = x;
	luacursory = y;
	setg3dcxy((float)x / screen->w, float(y) / screen->h);
	return true;
}

/**
 * This uses reference parameters to "return" the location
 * of the Lua UI cursor as a percentage of the screen width and height.
 */
void luaui_cursorpos(float &x, float &y)
{
	x = (float)((float)(luacursorx) / (float)(screen->w));
	y = (float)((float)(luacursory) / (float)(screen->h));
}

/**
 * This is called by the main rendering loop in rendergl.cpp in order to
 * render the Lua UI.
 */
void luaui_render(int w, int h)
{
    if(luaUIParent == NULL || (luauienabled == 0)) return;
	// glPushAttrib(GL_ENABLE_BIT); // scissor
	// glEnable(GL_SCISSOR_TEST);
	// glScissor(0, 0, 100, 100);
    luaUIParent->renderall(0,0);
	// glPopAttrib();
}

extern lua_State* getluastate();

// return the UIParent by lua value.
static int luaui_getuiparent(lua_State* L)
{
	lua_pushlightuserdata(L, (void*)&keyUiParent);
	lua_gettable(L, LUA_REGISTRYINDEX);
	return 1;
}
ADDLUAPROTECTEDAPI_N(GetUIParent, luaui_getuiparent);

static UIObject* findClicked(UIObject* parent)
{
	if(parent==NULL)
		return NULL;

	for(int i = parent->renderlist.size() - 1; i >= 0; --i)
	{
		// TODO: use non-1 return value to  initiate dragging or resize

		if( parent->renderlist[i]->containspoint(luacursorx, luacursory) != 0 && 
		  !parent->renderlist[i]->hidden )
		{
			return findClicked(parent->renderlist[i]);
		}
	}

	if( parent==luaUIParent )
		return NULL;

	return parent;
}

static void moveRenderToTop(UIObject* obj)
{
	if(obj == NULL)
		return;

	while(obj->parent != luaUIParent)
	{
		obj = obj->parent;
	}
	
	int index = -1;
	for(unsigned int i=0; i < luaUIParent->renderlist.size(); ++i)
	{
		if(luaUIParent->renderlist[i] == obj)
		{
			index = i;
			break;
		}
	}
	if(index < 0)
		return;

	if(luaUIParent->renderlist.size() > 0)
	{
		UIObject* swap = luaUIParent->renderlist[luaUIParent->renderlist.size()-1];
		luaUIParent->renderlist[luaUIParent->renderlist.size()-1] = obj;
		luaUIParent->renderlist[index] = swap;
	}
}

#define HANDLE_GUI_EVENT_ERROR(nargs, nrets) { \
	int l_err = lua_pcall(L, nargs, nrets, 0); \
	if(l_err != 0) { \
		conoutf("Lua UI error: %s", lua_tostring(L, -1)); \
		lua_pop(L, 1); \
	} \
}

void luaui_mousebutton(bool down, int btn)
{
	// TODO: handle button animation when they are implemented
	//
	if( luaUIParent == NULL || luauicursor == false) return;
	
	lua_State* L = getluastate();
	if( L == NULL ) return;

	// the object clicked becomes focus and receives mouse point
	if( down )
	{
		lua_pushlightuserdata(L, (void*)&keyFocus);
		UIObject* focusObj = findClicked(luaUIParent);
		classbind::pushobject(L, focusObj, "UIObject");
		if(focusObj != NULL)
			moveRenderToTop(focusObj);			
		lua_settable(L, LUA_REGISTRYINDEX);
	}
	
	lua_pushlightuserdata(L, (void*)&keyListeners);
	lua_gettable(L, LUA_REGISTRYINDEX);
	// send MouseDown to all registered listeners
	std::string evtName = down? "MouseDown" : "MouseUp";
	std::string delegateName = std::string("On") + evtName;
	
	UIObject* eachObject = NULL;
	lua_pushstring(L, evtName.c_str());
	lua_gettable(L, -2);
	lua_replace(L, -2);
	lua_pushnil(L);
	while(lua_next(L, -2))
	{
		eachObject = CLASSBIND_CHECKOBJECT(L, -2, UIObject);
		int pointIn = eachObject->containspoint(luacursorx, luacursory);
		if(pointIn == 1 && ! eachObject->hidden)
		{
			lua_pushstring(L, delegateName.c_str());
			lua_gettable(L, -3);
			if(!lua_isnil(L, -1))
			{
				lua_pushvalue(L, -1);
				lua_pushvalue(L, -4);
				lua_pushinteger(L, btn);
				HANDLE_GUI_EVENT_ERROR(2, 0);
			}
			lua_pop(L, 1);
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
}

void luaui_keypress(int sym, bool down)
{
	lua_State* L = getluastate();
	if( L == NULL ) return;
	
	// if an event registered KeyDown or KeyUp and is the focus object,
	// then call it's event handler
	std::string evtName = down? "KeyDown" : "KeyUp";
	std::string evtHandler = down? "OnKeyDown" : "OnKeyUp";

	lua_pushlightuserdata(L, (void*)&keyListeners);
	lua_gettable(L, LUA_REGISTRYINDEX); // listeners
	lua_pushstring(L, evtName.c_str());
	lua_gettable(L, -2);
	lua_replace(L, -2);
	lua_pushlightuserdata(L, (void*)&keyFocus);
	lua_gettable(L, LUA_REGISTRYINDEX); // listeners focus
	if(!lua_isnil(L, -1))
	{
		lua_gettable(L, -2); // listeners obj
		if(!lua_isnil(L, -1))
		{
			lua_getfield(L, -1, evtHandler.c_str());
			lua_pushvalue(L, -2);
			lua_pushinteger(L, sym);
			HANDLE_GUI_EVENT_ERROR(2, 1);
			// TODO: use a non-nil return value as sym when text fields are implemented
			// for text fields that handle this event.
			if(lua_isnil(L, -1))
			{
				if(lua_isboolean(L, -1))
				{
				}
				else if(lua_isnumber(L, -1))
				{
				}
				else if(lua_isstring(L, -1))
				{
				}
			}
			lua_pop(L, 1);
		}
		else
		{
			// printf("The focus object is not a keypress listener\n");
		}
	}
	else
	{
		// printf("The focus object is nil\n");
	}
	lua_pop(L, 2); // listeners, focus/obj
	
	// TODO: handle text fields when they are implemented
}

typedef UILayer UIParent;

static int lua_UIParent_mm__new(lua_State* L)
{
	UIParent* uiParent = new UIParent(NULL);
	lua_pushlightuserdata(L, uiParent);
	luaUIParent = uiParent;
	return 1;
}

static int lua_UIParent_mm__delete(lua_State* L)
{
	UIParent* uiParent = (UIParent*)lua_touserdata(L, 1);
	delete uiParent;
	luaUIParent = NULL;
	return 0;
}

static const luaL_Reg lua_UIParent_methods[] = {
	{NULL, NULL}
};

static const luaL_Reg lua_UIParent_metamethods[] = {
	{"__new", lua_UIParent_mm__new},
	{"__delete", lua_UIParent_mm__delete},
	{NULL, NULL}
};

/**
 * This is called in the Lua environment initialization function. It doesn't
 * initialize the UI classes; it initializes the UIParent object pointer and 
 * loads the blank texture.
 */
int luaopen_luaui(lua_State* L)
{
    if(blanktexture == NULL)
    {
        blanktexture = textureload("data/luaui/blank.png");
    }
    
    classbind::registerclass(L, "UIParent", lua_UIParent_methods, lua_UIParent_metamethods);
	CLASSBIND_ADDPARENT(L, UIParent, UIObject);
 
    lua_pushlightuserdata(L, (void*)&keyUiParent);
    lua_getglobal(L, "UIParent");
    lua_pushnil(L);
    lua_call(L, 1, 1);
    lua_settable(L, LUA_REGISTRYINDEX);
    luaUIParent->hidden = false;

	lua_pushnil(L);
	lua_setglobal(L, "UIParent"); // hide this constructor function

	lua_pushlightuserdata(L, (void*)&keyShownDlgs);
	NEW_WEAK_LIST(L, kv);
	lua_settable(L, LUA_REGISTRYINDEX);

	lua_pushlightuserdata(L, (void*)&keyButtons);
	NEW_WEAK_LIST(L, kv);
	lua_settable(L, LUA_REGISTRYINDEX);
	
	lua_pushlightuserdata(L, (void*)&keyListeners);
	lua_newtable(L);
	for(int i = 0; i < sizeof(guiEventNames) / sizeof(char*); ++i)
	{
		lua_pushstring(L, guiEventNames[i]);
		NEW_WEAK_LIST(L, kv);
		lua_settable(L, -3);
	}
	lua_settable(L, LUA_REGISTRYINDEX);

    return 0;
}


