/**
 @file
 @author darkside86-w00t
 
 Class object metatable layout =
 {
	magic = unique_number,
	classname = "Name of class",
	parents = { array_of_parent_metatables },
	downcast = { ["parent name"] = numbytes, ... },
	attributes = { table_of_int32_fields },
	get = { ["propname"]=accessor_func(key), ... },
	set = { ["propname"]=mutator_func(key,value) },
	luaobjects = { [ptr]=value, ... },
	cppobjects = { [ptr]=value, ... }
	
	-- optional metamethods:
	__new, __delete, __ctor, __dtor,
	
	-- methods that can be overloaded:
	__tostring,
 }
 
 Implementation of Lua classbind 0.3
**/

#include "luaclassbind.h"

static const char luaCbModule[] = {
#include "luacbmodule.hpp"
};

namespace classbind 
{
	//////
	// Class bind datatypes.
	//-------------------------------------------------------------------------
	
	struct ClassRef 
	{ 
		void *object;
		char needs_delete; // the luaobjects/cppobjects table is valid EXCEPT for __gc
	}; 
	
	/// primitive C++ datatypes
	enum CTYPES
	{
		CTYPE_VOID=0,
		CTYPE_INT8,
		CTYPE_UINT8,
		CTYPE_INT16,
		CTYPE_UINT16,
		CTYPE_INT32,
		CTYPE_UINT32,
		CTYPE_FLOAT,
		CTYPE_DOUBLE,
		CTYPE_STRING, // c string (readonly always)
	};
	
	//////
	// Registry values (a.k.a Lua "static globals")
	//-------------------------------------------------------------------------
	
	/// Holds the global table in which embedded Lua herein is executed
	static const char keyCbFenv = 'k';
	/// Indicates need to call initclassbind if nil
	static const char keyLibInit = 'k';
	/// Signature in metatable for all classrefs
	static const char keyClassRefSignature = 'k';
	
	static const luaL_Reg nomethods[] = { {NULL, NULL} };
	static const luaL_Reg nometamethods[] = { {NULL, NULL}};
	
	/////
	// Misc
	//-------------------------------------------------------------------------
	static void getLuaCbField(lua_State* L, const char* name)
	{
		lua_pushlightuserdata(L, (void*)&keyCbFenv);
		lua_gettable(L, LUA_REGISTRYINDEX);
		lua_pushstring(L, name);
		lua_gettable(L, -2);
		lua_replace(L, -2);
	}
	
	static void getClassName(lua_State* L, int index)
	{
		lua_getmetatable(L, index);
		lua_pushstring(L, "classname");
		lua_gettable(L, -2);
		lua_replace(L, -2);
	}
	
	static ClassRef* checkClassRef(lua_State* L, int index)
	{
		lua_pushvalue(L, index);
		lua_getmetatable(L, -1);
		lua_getfield(L, -1, "magic");
		lua_pushlightuserdata(L, (void*)&keyClassRefSignature);
		int equals = lua_equal(L, -1, -2);
		lua_pop(L, 4);
		if(equals == 0)
			luaL_typerror(L, index, "class object");
		return (ClassRef*)lua_touserdata(L, index);
	}
		
	//////
	// The library used by internal Lua code
	//-------------------------------------------------------------------------
	
	#define lua_pushint32(L, num) lua_pushinteger(L, static_cast<lua_Integer>( num ))
	
	static int luaCb_getcppfield(lua_State* L)
	{
		ClassRef* ref = checkClassRef(L, 1);
		lua_Integer field = lua_tointeger(L, 2);
		int type = (field & 0x7f000000) >> 24;
		char* addr = reinterpret_cast<char*>(
			reinterpret_cast<long long>(ref->object) +
			(0x00ffffff & field) 
		);
		switch(type)
		{
			case CTYPE_VOID:
				luaL_error(L, "Unknown C type");
				break;
			case CTYPE_INT8:
				lua_pushinteger(L, (signed)*reinterpret_cast<char*>(addr));
				break;
			case CTYPE_UINT8:
				lua_pushinteger(L, *reinterpret_cast<unsigned char*>(addr));
				break;
			case CTYPE_INT16:
				lua_pushinteger(L, (signed)*reinterpret_cast<short*>(addr));
				break;
			case CTYPE_UINT16:
				lua_pushinteger(L, *reinterpret_cast<unsigned short*>(addr));
				break;
			case CTYPE_INT32:
				lua_pushinteger(L, (signed)*reinterpret_cast<long*>(addr));
				break;
			case CTYPE_UINT32:
				lua_pushinteger(L, *reinterpret_cast<unsigned long*>(addr));
				break;
			case CTYPE_FLOAT:
				lua_pushnumber(L, *reinterpret_cast<float*>(addr));
				break;
			case CTYPE_DOUBLE: 
				lua_pushnumber(L, *reinterpret_cast<double*>(addr));
				break;
			case CTYPE_STRING:
				lua_pushstring(L, *reinterpret_cast<char**>(addr));
				break;
			default:
				luaL_error(L, "Unable to read C field info");
		}
		return 1;
	}
	
	static int luaCb_setcppfield(lua_State* L)
	{
		ClassRef* ref = checkClassRef(L, 1);
		lua_Integer field = lua_tointeger(L, 2);
		int type = (field & 0x7f000000) >> 24;
		void* addr = reinterpret_cast<void*>(
			reinterpret_cast<long long>(ref->object) +
			(0x00ffffff & field) 
		);
		bool isReadOnly = ((field & 0x80000000) >> 31) != 0 ? true : false;
		if(isReadOnly)
		{
			luaL_error(L, "Cannot set readonly field");
		}
		switch(type)
		{
			case CTYPE_VOID:
				luaL_error(L, "Unknown C type");
				break;
			case CTYPE_INT8:
				*reinterpret_cast<char*>(addr) = static_cast<char>(luaL_checkinteger(L,3));
				break;
			case CTYPE_UINT8:
				*reinterpret_cast<unsigned char*>(addr) = 
						static_cast<unsigned char>(luaL_checkinteger(L, 3));
				break;
			case CTYPE_INT16:
				*reinterpret_cast<short*>(addr) = static_cast<short>(luaL_checkinteger(L,3));
				break;
			case CTYPE_UINT16:
				*reinterpret_cast<unsigned short*>(addr) =
						static_cast<unsigned short>(luaL_checkinteger(L, 3));
				break;
			case CTYPE_INT32:
				*reinterpret_cast<long*>(addr) = static_cast<long>(luaL_checkinteger(L,3));
				break;
			case CTYPE_UINT32:
				*reinterpret_cast<unsigned long*>(addr) = 
						static_cast<unsigned long>(luaL_checkinteger(L,3));
				break;
			case CTYPE_FLOAT:
				*reinterpret_cast<float*>(addr) = static_cast<float>(luaL_checknumber(L,3));
				break;
			case CTYPE_DOUBLE:
				*reinterpret_cast<double*>(addr) = static_cast<double>(luaL_checknumber(L,3));
				break;
			case CTYPE_STRING:
				luaL_error(L, "Strings are readonly");
				break;
			default:
				luaL_error(L, "Unable to read field");
		}
		return 0;
	}
	
	static int luaCb_isclassref(lua_State* L)
	{
		lua_getmetatable(L, 1);
		lua_getfield(L, -1, "magic");
		lua_pushlightuserdata(L, (void*)&keyClassRefSignature);
		int isClassRef = lua_equal(L, -1, -2);
		lua_pop(L, 3);
		lua_pushboolean(L, isClassRef);
		return 1;
	}
	
	static int luaCb_getclassrefsig(lua_State* L)
	{
		lua_pushlightuserdata(L, (void*)&keyClassRefSignature);
		return 1;
	}
	
	static int luaCb_newclassref(lua_State* L)
	{
		void* obj = lua_touserdata(L, 1);
		ClassRef* ref = (ClassRef*)lua_newuserdata(L, sizeof(ClassRef));
		ref->object = obj;
		return 1;
	}
	
	static int luaCb_setobjfenv(lua_State* L)
	{
		lua_pushvalue(L, 2);
		lua_setfenv(L, 1);
		return 0;
	}
	
	static int luaCb_getobjfenv(lua_State* L)
	{
		lua_getfenv(L, 1);
		return 1;
	}
	
	// getregistry(index) - return a registry value by index
	static int luaCb_getregistry(lua_State* L)
	{
		lua_pushvalue(L, 1);
		lua_gettable(L, LUA_REGISTRYINDEX);
		return 1;
	}
	
	// unconditional metatable access
	static int luaCb_getmt(lua_State* L)
	{
		lua_getmetatable(L, 1);
		return 1;
	}
	
	static int luaCb_getobjclsname(lua_State* L)
	{
		lua_getmetatable(L, 1);
		lua_getfield(L, -1, "classname");
		lua_replace(L, -2);
		return 1;
	}
	
	static int luaCb_getobjptr(lua_State* L)
	{
		ClassRef* ref = (ClassRef*)lua_touserdata(L, 1);
		if(ref==NULL) return 0;
		lua_pushlightuserdata(L, ref->object);
		return 1;
	}
	
	static int luaCb_tointeger(lua_State* L)
	{
		void* ptr = lua_touserdata(L, 1);
		lua_pushinteger(L, reinterpret_cast<lua_Integer>(ptr));
		return 1;
	}
	
	static int luaCb_topointer(lua_State* L)
	{
		lua_Integer num = luaL_checkinteger(L, 1);
		lua_pushlightuserdata(L, reinterpret_cast<void*>(num));
		return 1;
	}

	static int luaCb_setneedsdelete(lua_State* L)
	{
		ClassRef* ref = (ClassRef*)lua_touserdata(L, 1);
		ref->needs_delete = lua_toboolean(L, 2);
		return 0;
	}

	static int luaCb_needsdelete(lua_State* L)
	{
		ClassRef* ref = (ClassRef*)lua_touserdata(L, 1);
		lua_pushboolean(L, ref->needs_delete);
		return 1;
	}
	
	static int luaCb_setmt(lua_State* L)
	{
		lua_pushvalue(L, 2);
		lua_setmetatable(L, 1);
		return 0;
	}
	
	static int luaCb_getlib(lua_State* L)
	{
		const char* name = luaL_checkstring(L, 1);
		lua_getglobal(L, name);
		return 1;
	}
	
	static int luaCb_newlib(lua_State* L)
	{
		const char* name = luaL_checkstring(L, 1);
		const luaL_Reg* methods = (luaL_Reg*)lua_touserdata(L, 2);
		luaL_register(L, name, methods);
		return 1;
	}
	
	static int luaCb_newmt(lua_State* L)
	{
		luaL_checkstring(L, 1);
		const luaL_Reg* metamethods = (luaL_Reg*)lua_touserdata(L, 2);
		int ok = luaL_newmetatable(L, lua_tostring(L, 1));
		if( ok == 0 )
		{
			lua_pushvalue(L, 1);
			lua_gettable(L, LUA_REGISTRYINDEX);
		}
		if(metamethods==NULL) metamethods = nometamethods;
		luaL_register(L, NULL, metamethods);
		return 1;
	}
	
	static const luaL_Reg luaCbFuncs[] = 
	{
		{"getcppfield", luaCb_getcppfield},
		{"setcppfield", luaCb_setcppfield},
		{"isclassref", luaCb_isclassref},
		{"getclassrefsig", luaCb_getclassrefsig},
		{"newclassref", luaCb_newclassref},
		{"setobjfenv", luaCb_setobjfenv},
		{"getobjfenv", luaCb_getobjfenv},
		{"getregistry", luaCb_getregistry},
		{"tointeger", luaCb_tointeger},
		{"topointer", luaCb_topointer},
		{"setneedsdelete", luaCb_setneedsdelete},
		{"needsdelete", luaCb_needsdelete},
		{"getmt", luaCb_getmt},
		{"setmt", luaCb_setmt},
		{"getobjclsname", luaCb_getobjclsname},
		{"getobjptr", luaCb_getobjptr},
		{"getlib", luaCb_getlib},
		{"newlib", luaCb_newlib},
		{"newmt", luaCb_newmt},
		{NULL, NULL},
	};
	
	/////
	// Initialization functions
	//-------------------------------------------------------------------------
	
	static int initLuaCbLib(lua_State* L)
	{
		// initialize the fenv of the internal lua code
		lua_pushlightuserdata(L, (void*)&keyCbFenv);
		lua_newtable(L);
		luaL_register(L, NULL, luaCbFuncs);
		// manually import some from global environment
		lua_pushstring(L, "sfmt");
		lua_getglobal(L, "string");
		lua_getfield(L, -1, "format");
		lua_replace(L, -2);
		lua_settable(L, -3);
		
		lua_pushstring(L, "error");
		lua_getglobal(L, "error");
		lua_settable(L, -3);
		
		lua_pushstring(L, "print");
		lua_getglobal(L, "print");
		lua_settable(L, -3);
		
		lua_pushstring(L, "ipairs");
		lua_getglobal(L, "ipairs");
		lua_settable(L, -3);
		
		lua_pushstring(L, "pcall");
		lua_getglobal(L, "pcall");
		lua_settable(L, -3);
		
		lua_pushstring(L, "ctypes");
		lua_newtable(L);
		#define SET_ENUM(name, value) { \
			lua_pushliteral(L, #name) ; lua_pushinteger(L, value) ; lua_settable(L, -3); \
		}
		SET_ENUM(int8, CTYPE_INT8);
		SET_ENUM(uint8, CTYPE_UINT8);
		SET_ENUM(int16, CTYPE_INT16);
		SET_ENUM(uint16, CTYPE_UINT16);
		SET_ENUM(int32, CTYPE_INT32);
		SET_ENUM(uint32, CTYPE_UINT32);
		SET_ENUM(float, CTYPE_FLOAT);
		SET_ENUM(double, CTYPE_DOUBLE);
		SET_ENUM(string, CTYPE_STRING);
		lua_settable(L, -3);
		
		lua_settable(L, LUA_REGISTRYINDEX);
		
		// run the lua cb module
		int notOk = luaL_loadbuffer(L, luaCbModule, sizeof(luaCbModule) -1, "lua classbind module");
		if(notOk)
		{
			printf("Failed to compile luaclassbind module: `%s'\n", lua_tostring(L, -1));
			lua_pop(L, 1);
		}
		lua_pushlightuserdata(L, (void*)&keyCbFenv);
		lua_gettable(L, LUA_REGISTRYINDEX);
		lua_setfenv(L, -2);
		notOk = lua_pcall(L, 0, 0, 0);
		if(notOk)
		{
			printf("Failed to run luaclassbind module: `%s'\n", lua_tostring(L, -1));
			lua_pop(L, 1);
		}
		
		return 0;
	}

	static int luatypeof(lua_State* L)
	{
		getLuaCbField(L, "typeof");
		lua_pushvalue(L, 1);
		lua_call(L, 1, 1);
		if(lua_isnil(L, -1))
			return luaL_typerror(L, 1, "class object");
		return 1;
	}
	
	/** initializes the library - not called manually, automatically done if
	* needed each time class is registered.
	*/
	static int initClassBind(lua_State* L)
	{
		initLuaCbLib(L);
		// initialize a few global utility methods
		lua_pushcfunction(L, luatypeof);
		lua_setglobal(L, "typeof");
		return 0;
	}
	
	//////
	// Public interface
	//-------------------------------------------------------------------------
	
	void registerclass(lua_State* L, const char* classname, const luaL_Reg* methods, 
					   const luaL_Reg* metamethods)
	{
		if(methods==NULL) methods=nomethods; if(metamethods==NULL) metamethods = nometamethods;
		lua_pushlightuserdata(L, (void*)&keyLibInit);
		lua_gettable(L, LUA_REGISTRYINDEX);
		bool libInit = !lua_isnil(L, -1);
		lua_pop(L, 1);
		if(!libInit)
		{
			initClassBind(L);
		}
		getLuaCbField(L, "registerclass");
		lua_pushstring(L, classname);
		lua_pushlightuserdata(L, (void*)methods);
		lua_pushlightuserdata(L, (void*)metamethods);
		lua_call(L, 3, 0);
	}
	
	void aliasclass(lua_State* L, const char* cls, const char* alias)
	{
		lua_pushstring(L, alias);
		lua_gettable(L, LUA_REGISTRYINDEX);
		if(!lua_isnil(L, -1))
			luaL_error(L, "Cannot overwrite existing class %s", alias);
		lua_pop(L, 1);
		
		lua_pushstring(L, alias);
		lua_pushstring(L, cls);
		lua_gettable(L, LUA_REGISTRYINDEX);
		lua_settable(L, LUA_REGISTRYINDEX);
	}
	
	void addparent(lua_State* L, const char* basename, const char* parentname, long long d)
	{
		getLuaCbField(L, "addparent");
		lua_pushstring(L, basename);
		lua_pushstring(L, parentname);
		lua_pushinteger(L, static_cast<lua_Integer>(d));
		lua_call(L, 3, 0);
	}
	
	void* checkobject_(lua_State* L, int index, const char* clsname)
	{
		lua_pushvalue(L, index);
		getLuaCbField(L, "checkobject");
		lua_insert(L, -2);
		lua_pushstring(L, clsname);
		lua_call(L, 2, 1);
		int ok = 0;
		void* result = NULL;
		
		if(lua_isnumber(L, -1))
			ok = lua_tointeger(L, -1);
		else if(lua_islightuserdata(L, -1))
			result = lua_touserdata(L, -1);
		lua_pop(L, 1);
		
		// an error?
		if(ok == -1)
		{
			luaL_typerror(L, index, "class object");
			return NULL;
		}
		else if(ok == -2)
		{
			getClassName(L, index);
			const char* argsclass = lua_tostring(L, -1);
			const char *msg = lua_pushfstring(L, "%s expected, got %s", clsname, argsclass);
			lua_pop(L, 1); // pop the wrong class's name
			luaL_argerror(L, index, msg);
			return NULL;
		}
		
		return result;
	}

	bool isobject(lua_State* L, int index, const char* classname)
	{
		lua_pushvalue(L, index);
		if(classname == NULL)
		{
			getLuaCbField(L, "isclassref");
			lua_insert(L, -2);
			lua_call(L, 1, 1);
			bool isRef = lua_toboolean(L, -1) != 0;
			lua_pop(L, 1);
			return isRef;
		}
		else
		{
			getLuaCbField(L, "checkobject");
			lua_insert(L, -2);
			lua_pushstring(L, classname);
			lua_call(L, 2, 1);
			bool isObj = false;
			if(lua_isnumber(L, -1))
				isObj = false;
			else if(lua_islightuserdata(L, -1))
				isObj = true;
			lua_pop(L, 1);
			return isObj;
		}
	}
	
	void pushobject(lua_State* L, void* object, const char* classname)
	{
		// if it is NULL push nil
		if(object==NULL)
		{
			lua_pushnil(L);
			return;
		}

		// if it is in luaobjects or cppobjects push by value
		// otherwise it is a C++ object
		getLuaCbField(L, "findobjectvalue");
		lua_pushstring(L, classname);
		lua_pushlightuserdata(L, object);
		lua_call(L, 2, 1);
		if(!lua_isnil(L, -1))
		{
			// object is now on stack
			return;
		}
		// otherwise, create new reference
		lua_pop(L, 1);
		getLuaCbField(L, "newcppobject");
		lua_pushstring(L, classname);
		lua_pushlightuserdata(L, object);
		lua_call(L, 2, 1);
	}

	void removeobject(lua_State* L, void* object, const char* classname)
	{
		getLuaCbField(L, "removeobject");
		lua_pushstring(L, classname);
		lua_pushlightuserdata(L, object);
		lua_call(L, 2, 0);
	}	

	void bindgetproperty(lua_State* L, const char* cls, const char* prop, lua_CFunction func)
	{
		getLuaCbField(L, "bindgetproperty");
		lua_pushstring(L, cls);
		lua_pushstring(L, prop);
		lua_pushcfunction(L, func);
		lua_call(L, 3, 0);
	}
	
	void bindsetproperty(lua_State* L, const char* cls, const char* prop, lua_CFunction func)
	{
		getLuaCbField(L, "bindsetproperty");
		lua_pushstring(L, cls);
		lua_pushstring(L, prop);
		lua_pushcfunction(L, func);
		lua_call(L, 3, 0);
	}
	
	void bindcppfield(lua_State* L, const char* cls, const char* name, const char* type, void* addr, 
					bool ro)
	{
		unsigned long a = reinterpret_cast<unsigned long>(addr);
		getLuaCbField(L, "ctypes");
		lua_pushstring(L, type);
		lua_gettable(L, -2);
		if(lua_isnil(L, -1))
			luaL_error(L, "Unknown C type %s", type);
		int t = lua_tointeger(L, -1);
		lua_pop(L, 2);
		int field = a & 0x00ffffff;
		field = field | ( (t&0x7f) << 24 );
		if (ro) field = field | (1 << 31);
		lua_pushstring(L, cls);
		lua_gettable(L, LUA_REGISTRYINDEX);
		if(lua_isnil(L, -1))
			luaL_error(L, "%s is not a registered class", cls);
		lua_getfield(L, -1, "fields");
		lua_pushstring(L, name);
		lua_pushinteger(L, field);
		lua_settable(L, -3);
		lua_pop(L, 2);
	}
}
