/** @file

@author darkside86-w00t

This library is intended to assist in binding C++ classes and full class
hierarchies to Lua. It does not automate "glue" code for functions, it
simply minimizes the amount of code needed to bind each class and maximizes
code reuse. All datatypes and algorithms of the library are implemented
purely in Lua or Lua opcodes (via Lua's C API function calls).

The library was designed with complex polymorphic and multi-inheritance
relations in mind. The ideal result is to allow bindings to classes without
any modification to the original C++ class declaration. The only
modification needed to original C++ methods is a C++ destructor for classes
that can be instantiated in C++ but accessed from Lua, but deletable from
C++ at anytime.

A class can bind metamethods whose names are special:
__new - Return a light userdata of just the allocated object pointer. If this is defined, the
		class name table can be used as a function to construct an object.
__delete - Given a light user data representing the reconstructed object, can deallocate it
		   according to how new allocated it.
__ctor - After a new Lua value is created, it can be passed to this function for direct
		 manipulation (i.e. setting its f-environment table or storing it else where).
__dtor - Usually not needed, the Lua value is passed before it is garbage collected.

Any attempt to define other metamethods may be ignored if the base object class
implements them and calls them for all lua classbind objects.

*/
#ifndef LUACLASSBIND_H
#define LUACLASSBIND_H

#include "lua.hpp"

/// The interface to Lua classbind
namespace classbind 
{
	
	void registerclass(lua_State*, const char* classname, const luaL_Reg* methods, 
					   const luaL_Reg* metamethods);
	
	#define CLASSBIND_REGISTERCLASS(L, Class) \
		classbind::registerclass( L, #Class , lua_##Class##_methods, lua_##Class##_metamethods)
					   
	void aliasclass(lua_State*, const char* cls, const char* alias);
	
	void addparent(lua_State*, const char* basename, const char* parentname, long long d);
	
	#define CLASSBIND_ADDPARENT(L, Base, Derived) \
		classbind::addparent(L, #Base, #Derived, classbind::downcastdata< Base , Derived >( ) )
	
	void* checkobject_(lua_State*, int index, const char* classname);

	bool isobject(lua_State*, int index, const char* classname=NULL);
	
	#define CLASSBIND_CHECKOBJECT(L, idx, cls) \
		( ( cls * ) classbind::checkobject_(L, idx, #cls) )
	
	void pushobject(lua_State*, void* object, const char* classname);

	void removeobject(lua_State*, void* object, const char* classname);
	
	void bindgetproperty(lua_State* L, const char* cls, const char* prop, lua_CFunction func);
	void bindsetproperty(lua_State* L, const char* cls, const char* prop, lua_CFunction func);
	
	void bindcppfield(lua_State* L, const char* cls, const char* name, const char* type, void* addr, 
					bool ro=true);
	#define CLASSBIND_BINDCPPFIELD(L, Class, Field, type, ro) { \
		Class* foo__ = reinterpret_cast< Class * >(0x013370); \
		classbind::bindcppfield(L, #Class, #Field, #type, reinterpret_cast<void*>( \
			(char*)( & foo__-> Field ) - (char*)(foo__) ), ro); }
	
	// last parameter of addparent
	template <class Derived, class Base>
	long long downcastdata()
	{
		Derived* derived = reinterpret_cast<Derived*>(0x0313370);
		void* ptr = derived;
		Base* b1 = static_cast<Base*>(static_cast<Derived*>(ptr));
		Base* b2 = static_cast<Base*>(ptr);
		long long diff = reinterpret_cast<long long>(b1) - reinterpret_cast<long long>(b2);
		// one would think the result is always 0, but that is just NOT the case...
		return diff;
	}

/*
	#define CLASSBIND_CLASSMETHOD( Class, method ) static int lua_##Class##_##method(lua_State* L) { \
		Class* self = ( Class * ) classbind::checkobject_(L, 1, #Class);
	#define CLASSBIND_ENDMETHOD }
	
	#define CLASSBIND_METHODLIST( Class ) static const lua_##Class##_methods[] = {
	#define CLASSBIND_METHOD( Class, name ) {#name, lua_##Class##_##name} ,
	#define CLASSBIND_ENDMETHODLIST {NULL, NULL} };
*/
	
}

#endif
